import copy
import logging
from abc import ABC, abstractmethod
from typing import List, Any, Dict
import re

import numpy as np
import pandas as pd
from sklearn import compose

from ..vector_model import VectorRegressionModel, VectorClassificationModel

log = logging.getLogger(__name__)


def createSkLearnModel(modelConstructor, modelArgs, outputTransformer=None):
    model = modelConstructor(**modelArgs)
    if outputTransformer is not None:
        model = compose.TransformedTargetRegressor(regressor=model, transformer=outputTransformer)
    return model


def strSkLearnModel(model):
    """
    Creates a cleaned string representation of the model with line breaks and indentations removed

    :param model: the sklearn model for which to generate the cleaned string representation
    :return: the string representation
    """
    return re.sub(r",\s*", ", ", str(model))


class AbstractSkLearnVectorRegressionModel(VectorRegressionModel, ABC):
    """
    Base class for models built upon scikit-learn's model implementations
    """
    log = log.getChild(__qualname__)

    def __init__(self, modelConstructor, **modelArgs):
        """
        :param modelConstructor: the sklearn model constructor
        :param modelArgs: arguments to be passed to the sklearn model constructor
        """
        super().__init__()
        self.sklearnInputTransformer = None
        self.sklearnOutputTransformer = None
        self.modelConstructor = modelConstructor
        self.modelArgs = modelArgs

    def _toStringExcludes(self) -> List[str]:
        return super()._toStringExcludes() + ["sklearnInputTransformer", "sklearnOutputTransformer", "modelConstructor", "modelArgs"]

    def withSkLearnInputTransformer(self, sklearnInputTransformer) -> __qualname__:
        """
        :param sklearnInputTransformer: an optional sklearn preprocessor for normalising/scaling inputs
        :return: self
        """
        self.sklearnInputTransformer = sklearnInputTransformer
        return self

    def withSkLearnOutputTransformer(self, sklearnOutputTransformer):
        """
        :param sklearnOutputTransformer: an optional sklearn preprocessor for normalising/scaling outputs
        :return: self
        """
        self.sklearnOutputTransformer = sklearnOutputTransformer
        return self

    def _transformInput(self, inputs: pd.DataFrame, fit=False) -> pd.DataFrame:
        if self.sklearnInputTransformer is None:
            return inputs
        else:
            inputValues = inputs.values
            shapeBefore = inputValues.shape
            if fit:
                inputValues = self.sklearnInputTransformer.fit_transform(inputValues)
            else:
                inputValues = self.sklearnInputTransformer.transform(inputValues)
            if inputValues.shape != shapeBefore:
                raise Exception("sklearnInputTransformer changed the shape of the input, which is unsupported. Consider using an a DFTSkLearnTransformer in inputTransformers instead.")
            return pd.DataFrame(inputValues, index=inputs.index, columns=inputs.columns)

    def _updateModelArgs(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        """
        Designed to be overridden in order to make input data-specific changes to modelArgs

        :param inputs: the training input data
        :param outputs: the training output data
        """
        pass

    def _fit(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        inputs = self._transformInput(inputs, fit=True)
        self._updateModelArgs(inputs, outputs)
        self._fitSkLearn(inputs, outputs)

    @abstractmethod
    def _fitSkLearn(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        pass

    def _predict(self, x: pd.DataFrame):
        inputs = self._transformInput(x)
        return self._predictSkLearn(inputs)

    @abstractmethod
    def _predictSkLearn(self, inputs: pd.DataFrame):
        pass


class AbstractSkLearnMultipleOneDimVectorRegressionModel(AbstractSkLearnVectorRegressionModel, ABC):
    """
    Base class for models which use several sklearn models of the same type with a single
    output dimension to create a multi-dimensional model (for the case where there is more than one output dimension)
    """
    def __init__(self, modelConstructor, **modelArgs):
        super().__init__(modelConstructor, **modelArgs)
        self.models = {}

    def _toStringExcludes(self) -> List[str]:
        return super()._toStringExcludes() + ["models"]

    def _toStringAdditionalEntries(self) -> Dict[str, Any]:
        d = super()._toStringAdditionalEntries()
        if len(self.models) > 0:
            d["model[0]"] = strSkLearnModel(next(iter(self.models.values())))
        else:
            d["modelConstructor"] = f"{self.modelConstructor.__name__}{self.modelArgs}"
        return d

    def _fitSkLearn(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        for predictedVarName in outputs.columns:
            log.info(f"Fitting model for output variable '{predictedVarName}'")
            model = createSkLearnModel(self.modelConstructor,
                    self.modelArgs,
                    outputTransformer=copy.deepcopy(self.sklearnOutputTransformer))
            model.fit(inputs, outputs[predictedVarName])
            self.models[predictedVarName] = model

    def _predictSkLearn(self, inputs: pd.DataFrame) -> pd.DataFrame:
        results = {}
        for varName in self.models:
            results[varName] = self.models[varName].predict(inputs)
        return pd.DataFrame(results)


class AbstractSkLearnMultiDimVectorRegressionModel(AbstractSkLearnVectorRegressionModel, ABC):
    """
    Base class for models which use a single sklearn model with multiple output dimensions to create the multi-dimensional model
    """
    def __init__(self, modelConstructor, **modelArgs):
        super().__init__(modelConstructor, **modelArgs)
        self.model = None

    def _toStringExcludes(self) -> List[str]:
        return super()._toStringExcludes() + ["model"]

    def _toStringAdditionalEntries(self) -> Dict[str, Any]:
        d = super()._toStringAdditionalEntries()
        if self.model is not None:
            d["model"] = strSkLearnModel(self.model)
        else:
            d["modelConstructor"] = f"{self.modelConstructor.__name__}{self.modelArgs}"
        return d

    def _fitSkLearn(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        if len(outputs.columns) > 1:
            log.info(f"Fitting a single multi-dimensional model for all {len(outputs.columns)} output dimensions")
        self.model = createSkLearnModel(self.modelConstructor, self.modelArgs, outputTransformer=self.sklearnOutputTransformer)
        outputValues = outputs.values
        if outputValues.shape[1] == 1:  # for 1D output, shape must be (numSamples,) rather than (numSamples, 1)
            outputValues = np.ravel(outputValues)
        self.model.fit(inputs, outputValues)

    def _predictSkLearn(self, inputs: pd.DataFrame) -> pd.DataFrame:
        Y = self.model.predict(inputs)
        return pd.DataFrame(Y, columns=self.getModelOutputVariableNames())


class AbstractSkLearnVectorClassificationModel(VectorClassificationModel, ABC):
    def __init__(self, modelConstructor, **modelArgs):
        """
        :param modelConstructor: the sklearn model constructor
        :param modelArgs: arguments to be passed to the sklearn model constructor
        """
        super().__init__()
        self.modelConstructor = modelConstructor
        self.sklearnInputTransformer = None
        self.sklearnOutputTransformer = None
        self.modelArgs = modelArgs
        self.model = None

    def _toStringExcludes(self) -> List[str]:
        return super()._toStringExcludes() + ["modelConstructor", "sklearnInputTransformer", "sklearnOutputTransformer",
                                              "modelArgs", "model"]

    def _toStringAdditionalEntries(self) -> Dict[str, Any]:
        d = super()._toStringAdditionalEntries()
        if self.model is None:
            d["modelConstructor"] = f"{self.modelConstructor.__name__}{self.modelArgs}"
        else:
            d["model"] = strSkLearnModel(self.model)
        return d

    def withSkLearnInputTransformer(self, sklearnInputTransformer) -> __qualname__:
        """
        :param sklearnInputTransformer: an optional sklearn preprocessor for normalising/scaling inputs
        :return: self
        """
        self.sklearnInputTransformer = sklearnInputTransformer
        return self

    def withSkLearnOutputTransformer(self, sklearnOutputTransformer):
        """
        :param sklearnOutputTransformer: an optional sklearn preprocessor for normalising/scaling outputs
        :return: self
        """
        self.sklearnOutputTransformer = sklearnOutputTransformer
        return self

    def _updateModelArgs(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        """
        Designed to be overridden in order to make input data-specific changes to modelArgs

        :param inputs: the training input data
        :param outputs: the training output data
        """
        pass

    def _fitClassifier(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        inputValues = self._transformInput(inputs, fit=True)
        self._updateModelArgs(inputs, outputs)
        self.model = createSkLearnModel(self.modelConstructor, self.modelArgs, self.sklearnOutputTransformer)
        log.info(f"Fitting sklearn classifier of type {self.model.__class__.__name__}")
        self.model.fit(inputValues, np.ravel(outputs.values))

    def _transformInput(self, inputs: pd.DataFrame, fit=False) -> np.ndarray:
        inputValues = inputs.values
        if self.sklearnInputTransformer is not None:
            if fit:
                inputValues = self.sklearnInputTransformer.fit_transform(inputValues)
            else:
                inputValues = self.sklearnInputTransformer.transform(inputValues)
        return inputValues

    def _predict(self, x: pd.DataFrame):
        inputValues = self._transformInput(x)
        Y = self.model.predict(inputValues)
        return pd.DataFrame(Y, columns=self._predictedVariableNames)

    def _predictClassProbabilities(self, x: pd.DataFrame):
        inputValues = self._transformInput(x)
        Y = self.model.predict_proba(inputValues)
        return pd.DataFrame(Y, columns=self._labels)

    def get_params(self, deep=True):
        return self.model.get_params(deep=deep)

    def set_params(self, **params):
        self.model.set_params(**params)




import logging
import os
import pickle
from typing import List, Dict, Any, Iterable

import joblib

log = logging.getLogger(__name__)


def loadPickle(path, backend="pickle"):
    with open(path, "rb") as f:
        if backend == "pickle":
            return pickle.load(f)
        elif backend == "joblib":
            return joblib.load(f)
        else:
            raise ValueError(f"Unknown backend '{backend}'")


def dumpPickle(obj, picklePath, backend="pickle", protocol=pickle.HIGHEST_PROTOCOL):
    dirName = os.path.dirname(picklePath)
    if dirName != "":
        os.makedirs(dirName, exist_ok=True)
    with open(picklePath, "wb") as f:
        if backend == "pickle":
            try:
                pickle.dump(obj, f, protocol=protocol)
            except AttributeError as e:
                failingPaths = PickleFailureDebugger.debugFailure(obj)
                raise AttributeError(f"Cannot pickle paths {failingPaths} of {obj}: {str(e)}")
        elif backend == "joblib":
            joblib.dump(obj, f, protocol=protocol)
        else:
            raise ValueError(f"Unknown backend '{backend}'")


class PickleFailureDebugger:
    """
    A collection of methods for testing whether objects can be pickled and logging useful infos in case they cannot
    """

    enabled = False  # global flag controlling the behaviour of logFailureIfEnabled

    @classmethod
    def _debugFailure(cls, obj, path, failures, handledObjectIds):
        if id(obj) in handledObjectIds:
            return
        handledObjectIds.add(id(obj))

        try:
            pickle.dumps(obj)
        except:
            # determine dictionary of children to investigate (if any)
            if hasattr(obj, '__dict__'):  # Because of strange behaviour of getstate, here try-except is used instead of if-else
                try:  # Because of strange behaviour of getattr(_, '__getstate__'), we here use try-except
                    d = obj.__getstate__()
                    if type(d) != dict:
                        d = {"state": d}
                except:
                    d = obj.__dict__
            elif type(obj) == dict:
                d = obj
            elif type(obj) in (list, tuple, set):
                d = dict(enumerate(obj))
            else:
                d = {}

            # recursively test children
            haveFailedChild = False
            for key, child in d.items():
                childPath = list(path) + [f"{key}[{child.__class__.__name__}]"]
                haveFailedChild = cls._debugFailure(child, childPath, failures, handledObjectIds) or haveFailedChild

            if not haveFailedChild:
                failures.append(path)

            return True
        else:
            return False

    @classmethod
    def debugFailure(cls, obj) -> List[str]:
        """
        Recursively tries to pickle the given object and returns a list of failed paths

        :param obj: the object for which to recursively test pickling
        :return: a list of object paths that failed to pickle
        """
        handledObjectIds = set()
        failures = []
        cls._debugFailure(obj, [obj.__class__.__name__], failures, handledObjectIds)
        return [".".join(l) for l in failures]

    @classmethod
    def logFailureIfEnabled(cls, obj, contextInfo: str = None):
        """
        If the class flag 'enabled' is set to true, the pickling of the given object is
        recursively tested and the results are logged at error level if there are problems and
        info level otherwise.
        If the flag is disabled, no action is taken.

        :param obj: the object for which to recursively test pickling
        :param contextInfo: optional additional string to be included in the log message
        """
        if cls.enabled:
            failures = cls.debugFailure(obj)
            prefix = f"Picklability analysis for {obj}"
            if contextInfo is not None:
                prefix += " (context: %s)" % contextInfo
            if len(failures) > 0:
                log.error(f"{prefix}: pickling would result in failures due to: {failures}")
            else:
                log.info(f"{prefix}: is picklable")


def setstate(cls, obj, state: Dict[str, Any], renamedProperties: Dict[str, str] = None, newOptionalProperties: List[str] = None,
        newDefaultProperties: Dict[str, Any] = None) -> None:
    """
    Helper function for safe implementations of __setstate__ in classes, which appropriately handles the cases where
    a parent class already implements __setstate__ and where it does not. Call this function whenever you would actually
    like to call the super-class' implementation.
    Unfortunately, __setstate__ is not implemented in object, rendering super().__setstate__(state) invalid in the general case.

    :param cls: the class in which you are implementing __setstate__
    :param obj: the instance of cls
    :param state: the state dictionary
    :param renamedProperties: a mapping from old property names to new property names
    :param newOptionalProperties: a list of names of new property names, which, if not present, shall be initialised with None
    :param newDefaultProperties: a dictionary mapping property names to their default values, which shall be added if they are not present
    """
    # handle new/changed properties
    if renamedProperties is not None:
        for mOld, mNew in renamedProperties.items():
            if mOld in state:
                state[mNew] = state[mOld]
                del state[mOld]
    if newOptionalProperties is not None:
        for mNew in newOptionalProperties:
            if mNew not in state:
                state[mNew] = None
    if newDefaultProperties is not None:
        for mNew, mValue in newDefaultProperties.items():
            if mNew not in state:
                state[mNew] = mValue
    # call super implementation, if any
    s = super(cls, obj)
    if hasattr(s, '__setstate__'):
        s.__setstate__(state)
    else:
        obj.__dict__ = state


def getstate(cls, obj, transientProperties: Iterable[str] = None, excludedProperties: Iterable[str] = None,
        overrideProperties: Dict[str, Any] = None) -> Dict[str, Any]:
    """
    Helper function for safe implementations of __getstate__ in classes, which appropriately handles the cases where
    a parent class already implements __getstate__ and where it does not. Call this function whenever you would actually
    like to call the super-class' implementation.
    Unfortunately, __getstate__ is not implemented in object, rendering super().__getstate__() invalid in the general case.

    :param cls: the class in which you are implementing __getstate__
    :param obj: the instance of cls
    :param transientProperties: transient properties which be set to None in serialisations
    :param excludedProperties: properties which shall be completely removed from serialisations
    :param overrideProperties: a mapping from property names to values specifying (new or existing) properties which are to be set;
        use this to set a fixed value for an existing property or to add a completely new property
    :return: the state dictionary, which may be modified by the receiver
    """
    s = super(cls, obj)
    if hasattr(s, '__getstate__'):
        d = s.__getstate__()
    else:
        d = obj.__dict__.copy()
    if transientProperties is not None:
        for p in transientProperties:
            if p in d:
                d[p] = None
    if excludedProperties is not None:
        for p in excludedProperties:
            if p in d:
                del d[p]
    if overrideProperties is not None:
        for k, v in overrideProperties.items():
            d[k] = v
    return d


# imports required for backward compatibility (with pickled objects)
from .torch_base import MCDropoutCapableNNModule
from .torch_models.mlp.mlp_modules import MultiLayerPerceptron
from .torch_models.lstnet.lstnet_modules import LSTNetwork
from .torch_models.residualffn.residualffn_modules import ResidualFeedForwardNetwork
from ..util import markUsed

markUsed(MCDropoutCapableNNModule, MultiLayerPerceptron, LSTNetwork, ResidualFeedForwardNetwork)


from typing import Union

from ..evaluation import RegressionEvaluationUtil
from ..evaluation.crossval import VectorModelCrossValidationData
from ..evaluation.eval_util import TEvalData, TCrossValData
from ..evaluation.evaluator import VectorModelEvaluationData
from . import TorchVectorRegressionModel


class TorchVectorRegressionModelEvaluationUtil(RegressionEvaluationUtil):

    def _createPlots(self, data: Union[TEvalData, TCrossValData], resultCollector: RegressionEvaluationUtil.ResultCollector, subtitle=None):
        super()._createPlots(data, resultCollector, subtitle)
        if isinstance(data, VectorModelEvaluationData):
            self._addLossProgressionPlotIfTorchVectorRegressionModel(data.model, "loss-progression", resultCollector)
        elif isinstance(data, VectorModelCrossValidationData):
            if data.trainedModels is not None:
                for i, model in enumerate(data.trainedModels, start=1):
                    self._addLossProgressionPlotIfTorchVectorRegressionModel(model, "loss-progression-{i}", resultCollector)

    @staticmethod
    def _addLossProgressionPlotIfTorchVectorRegressionModel(model, plotName, resultCollector):
        if isinstance(model, TorchVectorRegressionModel):
            resultCollector.addFigure(plotName, model.model.trainingInfo.plotAll())



import atexit
import logging as lg
from logging import *
import sys
import time
from typing import List, Tuple

import pandas as pd


log = getLogger(__name__)

LOG_DEFAULT_FORMAT = '%(levelname)-5s %(asctime)-15s %(name)s:%(funcName)s - %(message)s'


def removeLogHandlers():
    """
    Removes all current log handlers
    """
    logger = getLogger()
    while logger.hasHandlers():
        logger.removeHandler(logger.handlers[0])


def isLogHandlerActive(handler):
    return handler in getLogger().handlers


def configureLogging(format=LOG_DEFAULT_FORMAT, level=lg.DEBUG):
    removeLogHandlers()
    basicConfig(level=level, format=format, stream=sys.stdout)
    getLogger("matplotlib").setLevel(lg.INFO)
    getLogger("urllib3").setLevel(lg.INFO)
    getLogger("msal").setLevel(lg.INFO)
    pd.set_option('display.max_colwidth', 255)


_fileLoggerPaths: List[str] = []
_isAtExitReportFileLoggerRegistered = False


def _atExitReportFileLogger():
    for path in _fileLoggerPaths:
        print(f"A log file was saved to {path}")


def addFileLogger(path):
    global _isAtExitReportFileLoggerRegistered
    log.info(f"Logging to {path} ...")
    handler = FileHandler(path)
    handler.setFormatter(Formatter(LOG_DEFAULT_FORMAT))
    Logger.root.addHandler(handler)
    _fileLoggerPaths.append(path)
    if not _isAtExitReportFileLoggerRegistered:
        atexit.register(_atExitReportFileLogger)
        _isAtExitReportFileLoggerRegistered = True


class StopWatch:
    """
    Represents a stop watch for timing an execution. Constructing an instance starts the stopwatch.
    """
    def __init__(self):
        self.startTime = time.time()

    def restart(self):
        self.startTime = time.time()

    def getElapsedTimeSecs(self) -> float:
        return time.time() - self.startTime


class StopWatchManager:
    """
    A singleton which manages a pool of named stopwatches, such that executions to be timed by referring to a name only -
    without the need for a limited scope.
    """
    _instance = None

    @classmethod
    def getInstance(cls):
        if cls._instance is None:
            cls._instance = StopWatchManager(42)
        return cls._instance

    def __init__(self, secret):
        if secret != 42:
            raise Exception("Use only the singleton instance via getInstance")
        self._stopWatches = {}

    def start(self, name):
        self._stopWatches[name] = time.time()

    def stop(self, name) -> float:
        """
        :param name: the name of the time
        :return: the time that has passed in seconds
        """
        timePassedSecs = time.time() - self._stopWatches[name]
        del self._stopWatches[name]
        return timePassedSecs

    def isRunning(self, name):
        return name in self._stopWatches


class LogTime:
    def __init__(self, name):
        self.name = name
        self.startTime = None

    def start(self):
        self.startTime = time.time()

    def stop(self):
        log.info(f"{self.name} completed in {time.time()-self.startTime:.3f} seconds")

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def __enter__(self):
        return self.start()


import geopandas as gp
import logging
import numpy as np
from shapely.geometry import MultiPoint
from typing import Callable, Union, Iterable

from ..clustering.clustering_base import EuclideanClusterer
from ..clustering import SkLearnEuclideanClusterer
from ..clustering.sklearn_clustering import SkLearnClustererProtocol
from ..util.cache import LoadSaveInterface
from .coordinates import validateCoordinates, extractCoordinatesArray, TCoordinates, GeoDataFrameWrapper
from ..util.profiling import timed

log = logging.getLogger(__name__)


class CoordinateEuclideanClusterer(EuclideanClusterer, GeoDataFrameWrapper):
    """
    Wrapper around a clustering model. This class adds additional, geospatial-specific features to the provided
    clusterer

    :param clusterer: an instance of ClusteringModel
    """
    def __init__(self, clusterer: EuclideanClusterer):
        self.clusterer = clusterer
        super().__init__(noiseLabel=clusterer.noiseLabel,
            maxClusterSize=clusterer.maxClusterSize, minClusterSize=clusterer.minClusterSize)

    class Cluster(EuclideanClusterer.Cluster, GeoDataFrameWrapper, LoadSaveInterface):
        """
        Wrapper around a coordinates array

        :param coordinates:
        :param identifier:
        """

        def __init__(self, coordinates: np.ndarray, identifier: Union[str, int]):
            validateCoordinates(coordinates)
            super().__init__(coordinates, identifier)

        def toGeoDF(self, crs='epsg:3857'):
            """
            Export the cluster as a GeoDataFrame of length 1 with the cluster as an instance of
            MultiPoint and the identifier as index.

            :param crs: projection. By default pseudo-mercator
            :return: GeoDataFrame
            """
            gdf = gp.GeoDataFrame({"geometry": [self.asMultipoint()]}, index=[self.identifier])
            gdf.index.name = "identifier"
            gdf.crs = crs
            return gdf

        def asMultipoint(self):
            """
            :return: The cluster's coordinates as a MultiPoint object
            """
            return MultiPoint(self.datapoints)

        @classmethod
        def load(cls, path):
            """
            Instantiate from a geopandas readable file containing a single row with an identifier and an instance
            of MultiPoint

            :param path:
            :return: instance of CoordinateCluster
            """
            log.info(f"Loading instance of {cls.__name__} from {path}")
            gdf = gp.read_file(path)
            if len(gdf) != 1:
                raise Exception(f"Expected {path} to contain a single row, instead got {len(gdf)}")
            identifier, multipoint = gdf.identifier.values[0], gdf.geometry.values[0]
            return cls(np.array([[p.x, p.y] for p in multipoint]), identifier)

        def save(self, path, crs="EPSG:3857"):
            """
            Saves the cluster's coordinates as shapefile

            :param crs:
            :param path:
            :return:
            """
            log.info(f"Saving instance of {self.__class__.__name__} as shapefile to {path}")
            self.toGeoDF(crs).to_file(path, index=True)

    def _computeLabels(self, x: np.ndarray) -> np.ndarray:
        validateCoordinates(x)
        return self.clusterer._computeLabels(x)

    def fit(self, coordinates: TCoordinates):
        """
        Fitting to coordinates from a numpy array, a MultiPoint object or a GeoDataFrame with one Point per row

        :param coordinates:
        :return:
        """
        coordinates = extractCoordinatesArray(coordinates)
        super().fit(coordinates)

    @timed
    def toGeoDF(self, condition: Callable[[Cluster], bool] = None, crs='epsg:3857',
            includeNoise=False) -> gp.GeoDataFrame:
        """
        GeoDataFrame containing all clusters found by the model.
        It is a concatenation of GeoDataFrames of individual clusters

        :param condition: if provided, only clusters fulfilling the condition will be included
        :param crs: projection. By default pseudo-mercator
        :param includeNoise:
        :return: GeoDataFrame with all clusters indexed by their identifier
        """
        geodf = gp.GeoDataFrame()
        geodf.crs = crs
        # TODO or not TODO: parallelize this or improve performance some another way
        for cluster in self.clusters(condition):
            geodf = geodf.append(cluster.toGeoDF(crs=crs))
        if includeNoise:
            geodf = geodf.append(self.noiseCluster().toGeoDF(crs=crs))
        return geodf

    def plot(self, includeNoise=False, condition=None, **kwargs):
        """
        Plots the resulting clusters with random coloring

        :param includeNoise: Whether to include the noise cluster
        :param condition: If provided, only clusters fulfilling this condition will be included
        :param kwargs: passed to GeoDataFrame.plot
        :return:
        """
        geodf = self.toGeoDF(condition=condition, includeNoise=includeNoise)
        geodf["color"] = np.random.random(len(geodf))
        if includeNoise:
            geodf.loc[self.noiseLabel, "color"] = 0
        geodf.plot(column="color", **kwargs)

    # the overriding of the following methods is only necessary for getting the type annotations right
    # if mypy ever permits annotating nested classes correctly, these methods can be removed
    def getCluster(self, clusterId: int) -> Cluster:
        return super().getCluster(clusterId)

    def noiseCluster(self) -> Cluster:
        return super().noiseCluster()

    def clusters(self, condition: Callable[[Cluster], bool] = None) -> Iterable[Cluster]:
        return super().clusters(condition=condition)


class SkLearnCoordinateClustering(CoordinateEuclideanClusterer):
    """
    Wrapper around a sklearn clusterer. This class adds additional features like relabelling and convenient methods
    for handling geospatial data

    :param clusterer: a clusterer object compatible the sklearn API
    :param noiseLabel: label that is associated with the noise cluster or None
    :param minClusterSize: if not None, clusters below this size will be labeled as noise
    :param maxClusterSize: if not None, clusters above this size will be labeled as noise
    """
    def __init__(self, clusterer: SkLearnClustererProtocol, noiseLabel=-1,
                 minClusterSize: int = None, maxClusterSize: int = None):
        clusterer = SkLearnEuclideanClusterer(clusterer, noiseLabel=noiseLabel,
                           minClusterSize=minClusterSize, maxClusterSize=maxClusterSize)
        super().__init__(clusterer)



from __future__ import annotations
import collections
import functools
import pickle
import sys
from abc import ABC, abstractmethod
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Callable, Dict, Union, Any, List, Sequence, Optional
import json
import logging
import re
import threading


from azure.storage.table import TableService, TableBatch, Entity
from azure.storage.blob import BlockBlobService
import pandas as pd
import numpy as np

from .cache import PersistentKeyValueCache, PeriodicUpdateHook

_log = logging.getLogger(__name__)


class Serialiser(ABC):
    """
    Abstraction for mechanisms to serialise values, which do not fit table storage data model,
    see https://docs.microsoft.com/en-us/rest/api/storageservices/understanding-the-table-service-data-model
    """

    @abstractmethod
    def serialise(self, value) -> str:
        pass

    @abstractmethod
    def deSerialise(self, value: str):
        pass


class NumpyArrayJsonSerialiser(Serialiser):
    """
    Serialises a numpy array as json string of list representation of array
    """

    def serialise(self, value: np.ndarray) -> str:
        return json.dumps(value.tolist())

    def deSerialise(self, value: str):
        return np.array(json.loads(value))


class PropertyLoader(ABC):
    """
    Abstraction of a customised loader for an entity property
    """

    @abstractmethod
    def loadPropertyValue(self, entity: Entity):
        pass

    @abstractmethod
    def writePropertyValue(self, entity: Entity):
        pass

    @abstractmethod
    def loadPropertyValueToDataFrameColumn(self, df: pd.DataFrame):
        pass


class SerialisedPropertyLoader(PropertyLoader):
    """
    PropertyLoader to serialise and de-serialise values. Useful, if type of values is not aligned with table storage data model,
    see https://docs.microsoft.com/en-us/rest/api/storageservices/understanding-the-table-service-data-model
    """
    def __init__(self, propertyName: str, serialiser: Serialiser):
        self.serialiser = serialiser
        self.propertyName = propertyName

    def loadPropertyValue(self, entity: Entity):
        entity[self.propertyName] = self.serialiser.deSerialise(entity[self.propertyName])

    def writePropertyValue(self, entity: Entity):
        entity[self.propertyName] = self.serialiser.serialise(entity[self.propertyName])

    def loadPropertyValueToDataFrameColumn(self, df: pd.DataFrame):
        if self.propertyName in df.columns:
            df.loc[:, self.propertyName] = [self.serialiser.deSerialise(value) for value in df[self.propertyName]]


class AzureTableBlobBackend(ABC):
    """
    Abstraction of a blob backend, which allows for convenient setting and getting of values stored in blob storage via a
    reference to the value
    """

    @abstractmethod
    def getValueFromReference(self, valueIdentifier: str):
        pass

    @abstractmethod
    def getValueReference(self, partitionKey: str, rowKey: str, valueName: str, blobNamePrefix: str = None) -> str:
        pass

    @abstractmethod
    def setValueForReference(self, valueIdentifier: str, value):
        pass


class BlobPerKeyAzureTableBlobBackend(AzureTableBlobBackend, ABC):

    """
    Backend stores serialised values as /tableName/partitionKey/rowKey/valueName.<fileExtension>
    or /tableName/rowKey/valueName.<fileExtension>, if partitionKey equals tableName
    """

    def __init__(self, blockBlobService: BlockBlobService, containerName: str):
        """

        :param blockBlobService: https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.blockblobservice.blockblobservice?view=azure-python-previous
        """
        self.blockBlobService = blockBlobService
        self.containerName = containerName
        self.containerList = [container.name for container in blockBlobService.list_containers()]
        if containerName not in self.containerList:
            self.blockBlobService.create_container(containerName)
            self.containerList.append(containerName)

    @property
    @abstractmethod
    def fileExtension(self):
        pass

    @abstractmethod
    def _getBlobValue(self, containerName, blobName):
        pass

    @abstractmethod
    def _writeValueToBlob(self, containerName, blobName, value):
        pass

    def getValueFromReference(self, valueIdentifier: str):
        containerName = self._getContainerNameFromIdentifier(valueIdentifier)
        blobName = self._getBlobNameFromIdentifier(valueIdentifier)
        return self._getBlobValue(containerName, blobName)

    def getValueReference(self, partitionKey: str, rowKey: str, valueName: str, blobNamePrefix: str = None) -> str:
        blobName = self._getBlobNameFromKeys(partitionKey, rowKey, valueName, blobPrefix=blobNamePrefix)
        return self.blockBlobService.make_blob_url(self.containerName, blobName)

    def setValueForReference(self, valueIdentifier: str, value):
        containerName = self._getContainerNameFromIdentifier(valueIdentifier)
        blobName = self._getBlobNameFromIdentifier(valueIdentifier)
        self._writeValueToBlob(containerName, blobName, value)

    def _getBlobNameFromIdentifier(self, valueIdentifier: str):
        return (valueIdentifier.partition(f"{self.blockBlobService.primary_endpoint}/")[2]).partition("/")[2]

    def _getContainerNameFromIdentifier(self, valueIdentifier: str):
        return (valueIdentifier.partition(f"{self.blockBlobService.primary_endpoint}/")[2]).partition("/")[0]

    def _getBlobNameFromKeys(self, partitionKey: str, rowKey: str, valueName: str, blobPrefix: str = None):
        identifierList = [blobPrefix, partitionKey] if blobPrefix is not None and blobPrefix != partitionKey else [partitionKey]
        identifierList.extend([rowKey, valueName])
        return "/".join(identifierList) + self.fileExtension


class TextDumpAzureTableBlobBackend(BlobPerKeyAzureTableBlobBackend):
    """
   Backend stores values as txt files in the structure /tableName/partitionKey/rowKey/valueName
   """

    @property
    def fileExtension(self):
        return ""

    def _getBlobValue(self, containerName, blobName):
        return self.blockBlobService.get_blob_to_text(containerName, blobName).content

    def _writeValueToBlob(self, containerName, blobName, value):
        self.blockBlobService.create_blob_from_text(containerName, blobName, value)


class JsonAzureTableBlobBackend(BlobPerKeyAzureTableBlobBackend):
    """
    Backend stores values as json files in the structure /tableName/partitionKey/rowKey/valueName.json
    """

    @property
    def fileExtension(self):
        return ".json"

    def _getBlobValue(self, containerName, blobName):
        encodedValue = self.blockBlobService.get_blob_to_bytes(containerName, blobName).content
        return self._decodeBytesToValue(encodedValue)

    def _writeValueToBlob(self, containerName, blobName, value):
        encodedValue = self._encodeValueToBytes(value)
        self.blockBlobService.create_blob_from_bytes(containerName, blobName, encodedValue)

    @staticmethod
    def _encodeValueToBytes(value):
        return str.encode(json.dumps(value))

    @staticmethod
    def _decodeBytesToValue(_bytes):
        return json.loads(_bytes.decode())


class PickleAzureTableBlobBackend(JsonAzureTableBlobBackend):
    """
    Backend stores values as pickle files in the structure /tableName/partitionKey/rowKey/valueName.pickle
    """

    @property
    def fileExtension(self):
        return ".pickle"

    @staticmethod
    def _encodeValueToBytes(value):
        return pickle.dumps(value)

    @staticmethod
    def _decodeBytesToValue(_bytes):
        return pickle.loads(_bytes)


class BlobBackedPropertyLoader(PropertyLoader):
    AZURE_ALLOWED_SIZE_PER_PROPERTY_BYTES = 64000
    AZURE_ALLOWED_STRING_LENGTH_PER_PROPERTY = 32000

    """
    PropertyLoader to write and read values from blob backend via a reference to the value. Useful, if values cannot
    be stored in table storage itself, due to not being aligned with table storage data model, 
    see https://docs.microsoft.com/en-us/rest/api/storageservices/understanding-the-table-service-data-model
    """
    def __init__(self, propertyName: str, blobBackend: AzureTableBlobBackend, blobPrefix: str = None, propertyBooleanBlobStatusName: str = None, max_workers=None):
        """

        :param propertyName: name of property in table
        :param propertyBooleanBlobStatusName: name of property representing a boolean flag within a table, which indicates, if value is blob backed.
                                              If None, each value is assumed to be blob backed.
        :param blobBackend: actual backend to use for storage
        :param blobPrefix: prefix to use for blob in storage, e.g. a table name
        :param max_workers: maximal number of workers to load data from blob storage
        """
        self.blobPrefix = blobPrefix
        self.propertyBlobStatusName = propertyBooleanBlobStatusName
        self.blobBackend = blobBackend
        self.max_workers = max_workers
        self.propertyName = propertyName

    def loadPropertyValue(self, entity: Entity):
        if self._isEntityValueBlobBacked(entity):
            entity[self.propertyName] = self.blobBackend.getValueFromReference(entity[self.propertyName])

    def writePropertyValue(self, entity: Entity):
        if self.propertyName in entity.keys():
            if self._needToWriteToBlob(entity[self.propertyName]):
                valueIdentifier = self.blobBackend.getValueReference(entity["PartitionKey"], entity["RowKey"], self.propertyName, blobNamePrefix=self.blobPrefix)
                value = entity[self.propertyName]
                self.blobBackend.setValueForReference(valueIdentifier, value)
                entity[self.propertyName] = valueIdentifier
                propertyBlobStatus = True if self.propertyBlobStatusName is not None else None
            else:
                propertyBlobStatus = False if self.propertyBlobStatusName is not None else None

            if propertyBlobStatus is not None:
                entity[self.propertyBlobStatusName] = propertyBlobStatus

    def loadPropertyValueToDataFrameColumn(self, df: pd.DataFrame):
        if self.propertyName in df.columns:
            if self.propertyBlobStatusName is None:
                df.loc[:, self.propertyName] = self._loadValuesInSeries(df[self.propertyName])
            else:
                df.loc[df[self.propertyBlobStatusName], self.propertyName] = self._loadValuesInSeries(df.loc[df[self.propertyBlobStatusName], self.propertyName])

    def _needToWriteToBlob(self, value):
        if self.propertyBlobStatusName is None:
            return True
        if sys.getsizeof(value) > self.AZURE_ALLOWED_SIZE_PER_PROPERTY_BYTES:
            return True
        if isinstance(value, str) and len(value) > self.AZURE_ALLOWED_STRING_LENGTH_PER_PROPERTY:
            return True
        return False

    def _isEntityValueBlobBacked(self, entity: Entity):
        if self.propertyName not in entity.keys():
            return False
        if self.propertyBlobStatusName is None or self.propertyBlobStatusName not in entity:
            return True
        return entity[self.propertyBlobStatusName]

    def _loadValuesInSeries(self, _series: pd.Series):
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            _series = list(executor.map(self.blobBackend.getValueFromReference, _series))
        return _series


class BlobBackedSerialisedPropertyLoader(BlobBackedPropertyLoader, SerialisedPropertyLoader):
    """
    Property loader, which combines serialisation and blob backing.
    """
    def __init__(self, propertyName, serialiser: Serialiser, blobBackend: AzureTableBlobBackend, blobPrefix: str = None,
            propertyBooleanBlobStatusName: str = None, max_workers=None):
        """


        :param propertyName: name of property in table
        :param serialiser:
        :param propertyBooleanBlobStatusName: name of property representing a boolean flag within a table, which indicates, if value is blob backed.
                                              If None, each value is assumed to be blob backed.
        :param blobBackend: actual backend to use for storage
        :param blobPrefix: prefix to use for blob in storage, e.g. a table name
        :param max_workers: maximal number of workers to load data from blob storage
        """
        SerialisedPropertyLoader.__init__(self, propertyName, serialiser)
        BlobBackedPropertyLoader.__init__(self, propertyName, blobBackend, blobPrefix, propertyBooleanBlobStatusName, max_workers)

    def loadPropertyValue(self, entity: Entity):
        super(BlobBackedPropertyLoader, self).loadPropertyValue(entity)
        super(SerialisedPropertyLoader, self).loadPropertyValue(entity)

    def writePropertyValue(self, entity: Entity):
        super(SerialisedPropertyLoader, self).writePropertyValue(entity)
        super(BlobBackedPropertyLoader, self).writePropertyValue(entity)

    def loadPropertyValueToDataFrameColumn(self, df: pd.DataFrame):
        super(BlobBackedPropertyLoader, self).loadPropertyValueToDataFrameColumn(df)
        super(SerialisedPropertyLoader, self).loadPropertyValueToDataFrameColumn(df)


class AzureLazyBatchCommitTable:
    """
    Wrapper for an Azure table, which allow for convenient insertion via lazy batch execution per partition.
    Uses a priority queue to manage order of partitions to be committed.
    To execute insertions, call :func:`LazyBatchCommitTable.commit`
    """

    AZURE_ALLOWED_TABLE_NAME_PATTERN = re.compile("^[A-Za-z][A-Za-z0-9]{2,62}$")
    AZURE_ALLOWED_TABLE_BATCH_SIZE = 100

    class PartitionCommandsPriorityQueue:

        class PartitionCommands:
            def __init__(self, partitionKey):
                self.partitionKey = partitionKey
                self._commandList = collections.deque()

            def __len__(self):
                return len(self._commandList)

            def append(self, command):
                self._commandList.append(command)

            def execute(self, contextManager: Callable[[], TableBatch],  batchSize: int):
                while len(self._commandList) > 0:
                    _slice = [self._commandList.popleft() for _ in range(min(batchSize, len(self._commandList)))]
                    _log.info(f"Committing {len(_slice)} cache entries to the database")
                    with contextManager() as batch:
                        for command in _slice:
                            command(batch)

        def __init__(self):
            self.partitionCommandsQueue = []
            self.partitionKey2Commands = {}
            self._threadLock = threading.Lock()

        def addCommand(self, partitionKey, command: Union[Callable[[TableBatch], Any], functools.partial[TableBatch]]):
            """
            Add a command to queue of corresponding partitionKey
            :param partitionKey:
            :param command: a callable on a TableBatch
            """
            with self._threadLock:
                if partitionKey not in self.partitionKey2Commands:
                    commands = self.PartitionCommands(partitionKey)
                    self.partitionCommandsQueue.append(commands)
                    self.partitionKey2Commands[partitionKey] = commands
                self.partitionKey2Commands[partitionKey].append(command)

        def pop(self, minLength: int = None) -> Optional[AzureLazyBatchCommitTable.PartitionCommandsPriorityQueue.PartitionCommands]:
            """
            :param minLength: minimal length of largest PartitionCommands for the pop to take place.
            :return: largest PartitionCommands or None if minimal length is not reached
            """
            with self._threadLock:
                return self._pop(minLength)

        def popAll(self):
            with self._threadLock:
                commandsList = []
                while not self._isEmpty():
                    commandsList.append(self._pop())
                return commandsList

        def isEmpty(self):
            with self._threadLock:
                return self._isEmpty()

        def _pop(self, minLength=None):
            length, index = self._getMaxPriorityInfo()
            if index is not None and (minLength is None or length >= minLength):
                q = self.partitionCommandsQueue.pop(index)
                del self.partitionKey2Commands[q.partitionKey]
                return q
            else:
                return None

        def _isEmpty(self):
            return len(self.partitionCommandsQueue) == 0

        def _getMaxPriorityInfo(self):
            lengthsList = list(map(len, self.partitionCommandsQueue))
            if len(lengthsList) == 0:
                return 0, None
            maxLength = max(lengthsList)
            return maxLength, lengthsList.index(maxLength)

    def __init__(self, tableName: str, tableService: TableService, propertyLoaders: Sequence[PropertyLoader] = ()):
        """

        :param tableName: name of table
        :param tableService: instance of :class:`azure.storage.table.TableService` to connect to Azure table storage
        :param propertyLoaders:
        """

        if not self.AZURE_ALLOWED_TABLE_NAME_PATTERN.match(tableName):
            raise ValueError(f"Invalid table name {tableName}, see: https://docs.microsoft.com/en-us/rest/api/storageservices/Understanding-the-Table-Service-Data-Model")

        self.tableService = tableService
        self.tableName = tableName
        self.propertyLoaders = propertyLoaders
        self._partitionQueues = self.PartitionCommandsPriorityQueue()
        self._contextManager = functools.partial(self.tableService.batch, self.tableName)

        if not self.exists():
            self.tableService.create_table(self.tableName)

    def insertOrReplaceEntity(self, entity: Union[Dict, Entity]):
        """
        Lazy wrapper method for :func:`azure.storage.table.TableService.insert_or_replace_entity`
        :param entity:
        """
        partitionKey = entity["PartitionKey"]
        for propertyLoader in self.propertyLoaders:
            propertyLoader.writePropertyValue(entity)
        executionCommand = functools.partial(self._insertOrReplaceEntityViaBatch, entity)
        self._partitionQueues.addCommand(partitionKey, executionCommand)

    def insertEntity(self, entity: Union[Dict, Entity]):
        """
        Lazy wrapper method for :func:`azure.storage.table.TableService.insert_entity`
        :param entity:
        """
        partitionKey = entity["PartitionKey"]
        for propertyLoader in self.propertyLoaders:
            propertyLoader.writePropertyValue(entity)
        executionCommand = functools.partial(self._insertEntityViaBatch, entity)
        self._partitionQueues.addCommand(partitionKey, executionCommand)

    def getEntity(self, partitionKey: str, rowKey: str) -> Optional[Entity]:
        """
        Wraps :func:`azure.storage.table.TableService.get_entity`
        :param partitionKey:
        :param rowKey:
        :return:
        """
        try:
            entity = self.tableService.get_entity(self.tableName, partitionKey, rowKey)
            for propertyLoader in self.propertyLoaders:
                propertyLoader.loadPropertyValue(entity)
            return entity
        except Exception as e:
            _log.debug(f"Unable to load value for partitionKey {partitionKey} and rowKey {rowKey} from table {self.tableName}: {e}")
            return None

    def commitBlockingUntilEmpty(self, maxBatchSize=AZURE_ALLOWED_TABLE_BATCH_SIZE):
        """
        Commit insertion commands. Commands are executed batch-wise per partition until partition queue is empty in a
        blocking manner.
        :param maxBatchSize: maximal batch size to use for batch insertion, must be less or equal to batch size allowed by Azure
        """

        maxBatchSize = self._validateMaxBatchSize(maxBatchSize)

        while not self._partitionQueues.isEmpty():
            commands = self._partitionQueues.pop()
            commands.execute(self._contextManager, maxBatchSize)

    def commitNonBlockingCurrentQueueState(self, maxBatchSize=AZURE_ALLOWED_TABLE_BATCH_SIZE):
        """
        Commit insertion commands. Empties the current PartitionCommandsQueue in a non blocking way.
        Commands are executed batch-wise per partition.
        :param maxBatchSize: maximal batch size to use for batch insertion, must be less or equal to batch size allowed by Azure
        """

        maxBatchSize = self._validateMaxBatchSize(maxBatchSize)

        def commit():
            commandsList = self._partitionQueues.popAll()
            for commands in commandsList:
                commands.execute(self._contextManager, maxBatchSize)

        thread = threading.Thread(target=commit, daemon=False)
        thread.start()

    def commitBlockingLargestPartitionFromQueue(self, maxBatchSize=AZURE_ALLOWED_TABLE_BATCH_SIZE, minLength=None):
        """
        Commits in a blocking way the largest partition from PartitionCommandsQueue
        :param maxBatchSize: maximal batch size to use for batch insertion, must be less or equal to batch size allowed by Azure
        :param minLength: minimal size of largest partition. If not None, pop and commit only if minLength is reached.
        :return:
        """
        maxBatchSize = self._validateMaxBatchSize(maxBatchSize)
        commands = self._partitionQueues.pop(minLength)
        if commands is not None:
            commands.execute(self._contextManager, maxBatchSize)

    def _validateMaxBatchSize(self, maxBatchSize):
        if maxBatchSize > self.AZURE_ALLOWED_TABLE_BATCH_SIZE:
            _log.warning(f"Provided maxBatchSize is larger than allowed size {self.AZURE_ALLOWED_TABLE_BATCH_SIZE}. Will use maxBatchSize {self.AZURE_ALLOWED_TABLE_BATCH_SIZE} instead.")
            maxBatchSize = self.AZURE_ALLOWED_TABLE_BATCH_SIZE
        return maxBatchSize

    def loadTableToDataFrame(self, columns: List[str] = None, rowFilterQuery: str = None, numRecords: int = None):
        """
        Load all rows of table to :class:`~pandas.DataFrame`
        :param rowFilterQuery:
        :param numRecords:
        :param columns: restrict loading to provided columns
        :return: :class:`~pandas.DataFrame`
        """
        if numRecords is None:
            records = list(self._iterRecords(columns, rowFilterQuery))
        else:
            records = []
            for record in self._iterRecords(columns, rowFilterQuery):
                records.append(record)
                if len(records) >= numRecords:
                    break
        df = pd.DataFrame(records, columns=columns)
        for propertyLoader in self.propertyLoaders:
            propertyLoader.loadPropertyValueToDataFrameColumn(df)
        return df

    def iterDataFrameChunks(self, chunkSize: int, columns: List[str] = None, rowFilterQuery: str = None):
        """
        Get a generator of dataframe chunks
        :param rowFilterQuery:
        :param chunkSize:
        :param columns:
        :return:
        """
        records = []
        for record in self._iterRecords(columns, rowFilterQuery):
            records.append(record)
            if len(records) >= chunkSize:
                df = pd.DataFrame(records, columns=columns)
                for propertyLoader in self.propertyLoaders:
                    propertyLoader.loadPropertyValueToDataFrameColumn(df)
                yield df
                records = []

    def iterRecords(self, columns: List[str] = None, rowFilterQuery: str = None):
        """

        Get a generator of table entities
        :param rowFilterQuery:
        :param columns:
        :return:
        """
        for entity in self._iterRecords(columns, rowFilterQuery):
            for propertyLoader in self.propertyLoaders:
                propertyLoader.loadPropertyValue(entity)
            yield entity

    def _iterRecords(self, columns: Optional[List[str]], rowFilterQuery: Optional[str]):

        columnNamesAsCommaSeparatedString = None
        if columns is not None:
            columnNamesAsCommaSeparatedString = ",".join(columns)
        return self.tableService.query_entities(self.tableName, select=columnNamesAsCommaSeparatedString,
                filter=rowFilterQuery)

    def insertDataFrameToTable(self, df: pd.DataFrame, partitionKeyGenerator: Callable[[str], str] = None, numRecords: int = None):
        """
        Inserts or replace entities of the table corresponding to rows of the DataFrame, where the index of the dataFrame acts as rowKey.
        Values of object type columns in the dataFrame may have to be serialised via json beforehand.
        :param df: DataFrame to be inserted
        :param partitionKeyGenerator: if None, partitionKeys default to tableName
        :param numRecords: restrict insertion to first numRecords rows, merely for testing
        """
        for (count, (idx, row)) in enumerate(df.iterrows()):
            if numRecords is not None:
                if count >= numRecords:
                    break
            entity = row.to_dict()
            entity["RowKey"] = idx
            entity["PartitionKey"] = self.tableName if partitionKeyGenerator is None else partitionKeyGenerator(idx)
            self.insertOrReplaceEntity(entity)

    @staticmethod
    def _insertOrReplaceEntityViaBatch(entity, batch: TableBatch):
        return batch.insert_or_replace_entity(entity)

    @staticmethod
    def _insertEntityViaBatch(entity, batch: TableBatch):
        return batch.insert_entity(entity)

    def exists(self):
        return self.tableService.exists(self.tableName)


class AzureTablePersistentKeyValueCache(PersistentKeyValueCache):
    """
    PersistentKeyValueCache using Azure Table Storage, see https://docs.microsoft.com/en-gb/azure/storage/tables/
    """
    CACHE_VALUE_IDENTIFIER = "cache_value"

    def __init__(self, tableService: TableService, tableName="cache", partitionKeyGenerator: Callable[[str], str] = None,
            maxBatchSize=100, minSizeForPeriodicCommit: Optional[int] = 100, deferredCommitDelaySecs=1.0, inMemory=False,
            blobBackend: AzureTableBlobBackend = None, serialiser: Serialiser = None, max_workers: int = None):
        """


        :param tableService: https://docs.microsoft.com/en-us/python/api/azure-cosmosdb-table/azure.cosmosdb.table.tableservice.tableservice?view=azure-python
        :param tableName: name of table, needs to match restrictions for Azure storage resources, see https://docs.microsoft.com/en-gb/azure/azure-resource-manager/management/resource-name-rules
        :param partitionKeyGenerator: callable to generate a partitionKey from provided string, if None partitionKey in requests defaults to tableName
        :param maxBatchSize: maximal batch size for each commit.
        :param deferredCommitDelaySecs: the time frame during which no new data must be added for a pending transaction to be committed
        :param minSizeForPeriodicCommit: minimal size of a batch to be committed in a periodic thread.
                                         If None, commits are only executed in a deferred manner, i.e. commit only if there is no update for deferredCommitDelaySecs
        :param inMemory: boolean flag, to indicate, if table should be loaded in memory at construction
        :param blobBackend: if not None, blob storage will be used to store actual value and cache_value in table only contains a reference
        :param max_workers: maximal number of workers to load data from blob backend
        """

        self._deferredCommitDelaySecs = deferredCommitDelaySecs
        self._partitionKeyGenerator = partitionKeyGenerator

        def createPropertyLoaders():
            if blobBackend is None and serialiser is None:
                _propertyLoaders = ()
            elif blobBackend is None and serialiser is not None:
                _propertyLoaders = (SerialisedPropertyLoader(self.CACHE_VALUE_IDENTIFIER, serialiser),)
            elif blobBackend is not None and serialiser is None:
                propertyBlobStatusName = self.CACHE_VALUE_IDENTIFIER + "_blob_backed"
                _propertyLoaders = (BlobBackedPropertyLoader(self.CACHE_VALUE_IDENTIFIER, blobBackend, tableName,
                    propertyBlobStatusName, max_workers),)
            else:
                propertyBlobStatusName = self.CACHE_VALUE_IDENTIFIER + "_blob_backed"
                _propertyLoaders = (BlobBackedSerialisedPropertyLoader(self.CACHE_VALUE_IDENTIFIER, serialiser, blobBackend,
                tableName, propertyBlobStatusName, max_workers),)
            return _propertyLoaders

        propertyLoaders = createPropertyLoaders()
        self._batchCommitTable = AzureLazyBatchCommitTable(tableName, tableService, propertyLoaders=propertyLoaders)
        self._minSizeForPeriodicCommit = minSizeForPeriodicCommit
        self._maxBatchSize = maxBatchSize
        self._updateHook = PeriodicUpdateHook(deferredCommitDelaySecs, noUpdateFn=self._commit, periodicFn=self._periodicallyCommit)

        self._inMemoryCache = None

        if inMemory:
            df = self._batchCommitTable.loadTableToDataFrame(columns=['RowKey', self.CACHE_VALUE_IDENTIFIER]).set_index("RowKey")
            _log.info(f"Loaded {len(df)} entries of table {tableName} in memory")
            self._inMemoryCache = df[self.CACHE_VALUE_IDENTIFIER].to_dict()

    def set(self, key, value):
        keyAsString = str(key)
        partitionKey = self._getPartitionKeyForRowKey(keyAsString)
        entity = {'PartitionKey': partitionKey, 'RowKey': keyAsString, self.CACHE_VALUE_IDENTIFIER: value}
        self._batchCommitTable.insertOrReplaceEntity(entity)
        self._updateHook.handleUpdate()

        if self._inMemoryCache is not None:
            self._inMemoryCache[keyAsString] = value

    def get(self, key):
        keyAsString = str(key)
        value = self._getFromInMemoryCache(keyAsString)
        if value is None:
            value = self._getFromTable(keyAsString)
        return value

    def _getFromTable(self, key: str):
        partitionKey = self._getPartitionKeyForRowKey(key)
        entity = self._batchCommitTable.getEntity(partitionKey, key)
        if entity is not None:
            return entity[self.CACHE_VALUE_IDENTIFIER]
        return None

    def _getFromInMemoryCache(self, key):
        if self._inMemoryCache is None:
            return None
        return self._inMemoryCache.get(str(key), None)

    def _getPartitionKeyForRowKey(self, key: str):
        return self._batchCommitTable.tableName if self._partitionKeyGenerator is None else self._partitionKeyGenerator(key)

    def _commit(self):
        self._batchCommitTable.commitNonBlockingCurrentQueueState(self._maxBatchSize)

    def _periodicallyCommit(self):
        self._batchCommitTable.commitBlockingLargestPartitionFromQueue(self._maxBatchSize, self._minSizeForPeriodicCommit)



from . import torch_modules as modules, torch_models as models
from .torch_data import TensorScaler, DataUtil, VectorDataUtil
from .torch_opt import NNLossEvaluatorRegression, NNLossEvaluator, NNOptimiser
from .torch_base import TorchModel, VectorTorchModel, TorchVectorRegressionModel, \
    TorchVectorClassificationModel



import pandas as pd


def tsNextMonth(ts: pd.Timestamp) -> pd.Timestamp:
    m = ts.month
    if m == 12:
        return ts.replace(year=ts.year+1, month=1)
    else:
        return ts.replace(month=m+1)



from enum import Enum
from typing import Sequence, Optional, TypeVar, Generic, Tuple, Dict, Any

import pandas as pd

from . import sequences as array_util
from .string import ToStringMixin, dictString

TKey = TypeVar("TKey")
TValue = TypeVar("TValue")


class Trivalent(Enum):
    TRUE = "true"
    FALSE = "false"
    UNKNOWN = "unknown"

    @classmethod
    def fromBool(cls, b: bool):
        return cls.TRUE if b else cls.FALSE

    def isTrue(self):
        return self == Trivalent.TRUE

    def isFalse(self):
        return self == Trivalent.FALSE


class DeferredParams(ToStringMixin):
    """
    Represents a dictionary of parameters that is specifically designed to hold parameters that can only defined late within
    a process (i.e. not initially at construction time), e.g. because the parameters are data-dependent and therefore can only
    be determined once the data has been seen.
    """
    UNDEFINED = "__undefined__DeferredParams"

    def __init__(self):
        self.params = {}

    def _toStringObjectInfo(self) -> str:
        return dictString(self.params)

    def setParam(self, name: str, value: Any):
        self.params[name] = value

    def getParam(self, name, default=UNDEFINED):
        """
        :param name: the parameter name
        :param default: in case no value is set, return this value, and if UNDEFINED (default), raise KeyError
        :return: the parameter value
        """
        if default == self.UNDEFINED:
            return self.params[name]
        else:
            return self.params.get(name, default)

    def getDict(self) -> Dict[str, Any]:
        return self.params


class SortedValues(Generic[TValue]):
    """
    Provides convenient binary search (bisection) operations for sorted sequences
    """
    def __init__(self, sortedValues: Sequence[TValue]):
        self.values = sortedValues

    def __len__(self):
        return len(self.values)

    def floorIndex(self, value) -> Optional[int]:
        """
        Finds the rightmost index where the value is less than or equal to the given value

        :param value: the value to search for
        :return: the index or None if there is no such index
        """
        return array_util.floorIndex(self.values, value)

    def ceilIndex(self, value) -> Optional[int]:
        """
        Finds the leftmost index where the value is greater than or equal to the given value

        :param value: the value to search for
        :return: the index or None if there is no such index
        """
        return array_util.ceilIndex(self.values, value)

    def closestIndex(self, value) -> Optional[int]:
        """
        Finds the index of the value that is closest to the given value.
        If two subsequent values have the same distance, the smaller index is returned.

        :param value: the value to search for
        :return: the index or None if this object is empty
        """
        return array_util.closestIndex(self.values, value)

    def _value(self, idx: Optional[int]) -> Optional[TValue]:
        if idx is None:
            return None
        else:
            return self.values[idx]

    def floorValue(self, value) -> Optional[TValue]:
        """
        Finds the largest value that is less than or equal to the given value

        :param value: the value to search for
        :return: the value or None if there is no such value
        """
        return self._value(self.floorIndex(value))

    def ceilValue(self, value) -> Optional[TValue]:
        """
        Finds the smallest value that is greater than or equal to the given value

        :param value: the value to search for
        :return: the value or None if there is no such value
        """
        return self._value(self.ceilIndex(value))

    def closestValue(self, value) -> Optional[TValue]:
        """
        Finds the value that is closest to the given value.
        If two subsequent values have the same distance, the smaller value is returned.

        :param value: the value to search for
        :return: the value or None if this object is empty
        """
        return self._value(self.closestIndex(value))

    def _valueSlice(self, firstIndex, lastIndex):
        if firstIndex is None or lastIndex is None:
            return None
        return self.values[firstIndex:lastIndex+1]

    def valueSlice(self, lowestKey, highestKey) -> Optional[Sequence[TValue]]:
        return self._valueSlice(self.ceilIndex(lowestKey), self.floorIndex(highestKey))


class SortedKeysAndValues(Generic[TKey, TValue]):
    def __init__(self, keys: Sequence[TKey], values: Sequence[TValue]):
        """
        :param keys: a sorted sequence of keys
        :param values: a sequence of corresponding values
        """
        if len(keys) != len(values):
            raise ValueError("Lengths of keys and values do not match")
        self.keys = keys
        self.values = values

    @classmethod
    def fromSeries(cls, s: pd.Series):
        """
        Creates an instance from a pandas Series, using the series' index as the keys and its values as the values

        :param s: the series
        :return: an instance
        """
        return cls(s.index, s.values)

    def floorIndex(self, key) -> Optional[int]:
        """
        Finds the rightmost index where the key value is less than or equal to the given value

        :param key: the value to search for
        :return: the index or None if there is no such index
        """
        return array_util.floorIndex(self.keys, key)

    def ceilIndex(self, key) -> Optional[int]:
        """
        Finds the leftmost index where the key value is greater than or equal to the given value

        :param key: the value to search for
        :return: the index or None if there is no such index
        """
        return array_util.ceilIndex(self.keys, key)

    def closestIndex(self, key) -> Optional[int]:
        """
        Finds the index where the key is closest to the given value.
        If two subsequent keys have the same distance, the smaller index is returned.

        :param key: the value to search for
        :return: the index or None if this object is empty.
        """
        return array_util.closestIndex(self.keys, key)

    def floorValue(self, key) -> Optional[TValue]:
        """
        Returns the value for the largest index where the corresponding key is less than or equal to the given value

        :param value: the value to search for
        :return: the value or None if there is no such value
        """
        return array_util.floorValue(self.keys, key, values=self.values)

    def ceilValue(self, key) -> Optional[TValue]:
        """
        Returns the value for the smallest index where the corresponding key is greater than or equal to the given value

        :param value: the value to search for
        :return: the value or None if there is no such value
        """
        return array_util.ceilValue(self.keys, key, values=self.values)

    def closestValue(self, key) -> Optional[TValue]:
        """
        Finds the value that is closest to the given value.
        If two subsequent values have the same distance, the smaller value is returned.

        :param value: the value to search for
        :return: the value or None if this object is empty
        """
        return array_util.closestValue(self.keys, key, values=self.values)

    def floorKeyAndValue(self, key) -> Optional[Tuple[TKey, TValue]]:
        idx = self.floorIndex(key)
        return None if idx is None else (self.keys[idx], self.values[idx])

    def ceilKeyAndValue(self, key) -> Optional[Tuple[TKey, TValue]]:
        idx = self.ceilIndex(key)
        return None if idx is None else (self.keys[idx], self.values[idx])

    def closestKeyAndValue(self, key) -> Optional[Tuple[TKey, TValue]]:
        idx = self.closestIndex(key)
        return None if idx is None else (self.keys[idx], self.values[idx])

    def valueSliceInner(self, lowerBoundKey, upperBoundKey):
        return array_util.valueSliceOuter(self.keys, lowerBoundKey, upperBoundKey, values=self.values)

    def valueSliceOuter(self, lowerBoundKey, upperBoundKey):
        return array_util.valueSliceOuter(self.keys, lowerBoundKey, upperBoundKey, values=self.values)


class SortedKeyValuePairs(Generic[TKey, TValue]):
    @classmethod
    def fromUnsortedKeyValuePairs(cls, unsortedKeyValuePairs: Sequence[Tuple[TKey, TValue]]):
        return cls(sorted(unsortedKeyValuePairs, key=lambda x: x[0]))

    def __init__(self, sortedKeyValuePairs: Sequence[Tuple[TKey, TValue]]):
        self.entries = sortedKeyValuePairs
        self._sortedKeys = SortedValues([t[0] for t in sortedKeyValuePairs])

    def _value(self, idx: Optional[int]) -> Optional[TValue]:
        if idx is None:
            return None
        return self.entries[idx][1]

    def floorIndex(self, key) -> Optional[int]:
        """Finds the rightmost index where the key is less than or equal to the given key"""
        return self._sortedKeys.floorIndex(key)

    def floorValue(self, key) -> Optional[TValue]:
        return self._value(self.floorIndex(key))

    def floorKeyAndValue(self, key) -> Optional[Tuple[TKey, TValue]]:
        idx = self.floorIndex(key)
        return None if idx is None else self.entries[idx]

    def ceilIndex(self, key) -> Optional[int]:
        """Find leftmost index where the key is greater than or equal to the given key"""
        return self._sortedKeys.ceilIndex(key)

    def ceilValue(self, key) -> Optional[TValue]:
        return self._value(self.ceilIndex(key))

    def ceilKeyAndValue(self, key) -> Optional[Tuple[TKey, TValue]]:
        idx = self.ceilIndex(key)
        return None if idx is None else self.entries[idx]

    def closestIndex(self, key) -> Optional[int]:
        return self._sortedKeys.closestIndex(key)

    def closestValue(self, key) -> Optional[TValue]:
        return self._value(self.closestIndex(key))

    def closestKeyAndValue(self, key):
        idx = self.closestIndex(key)
        return None if idx is None else self.entries[idx]

    def _valueSlice(self, firstIndex, lastIndex):
        if firstIndex is None or lastIndex is None:
            return None
        return [e[1] for e in self.entries[firstIndex:lastIndex+1]]

    def valueSlice(self, lowestKey, highestKey) -> Optional[Sequence[TValue]]:
        return self._valueSlice(self.ceilIndex(lowestKey), self.floorIndex(highestKey))

    def slice(self, lowerBoundKey=None, upperBoundKey=None, inner=True) -> "SortedKeyValuePairs":
        """
        :param lowerBoundKey: the key defining the start of the slice (depending on inner);
            if None, the first included entry will be the very first entry
        :param upperBoundKey: the key defining the end of the slice (depending on inner);
            if None, the last included entry will be the very last entry
        :param inner: if True, the returned slice will be within the bounds; if False, the the returned
            slice is extended by one entry in both directions such that it contains the bounds (where possible)
        :return:
        """
        if lowerBoundKey is not None:
            fromIndex = self.ceilIndex(lowerBoundKey) if inner else self.floorIndex(lowerBoundKey)
            if fromIndex is None:
                fromIndex = 0
        else:
            fromIndex = 0
        if upperBoundKey is not None:
            toIndex = self.floorIndex(upperBoundKey) if inner else self.ceilIndex(upperBoundKey)
            if toIndex is None:
                toIndex = len(self.entries)
        else:
            toIndex = len(self.entries)
        return SortedKeyValuePairs(self.entries[fromIndex:toIndex])


import logging
import numpy as np
import seaborn as sns
from abc import abstractmethod, ABC
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from typing import List, Sequence, Optional

from .eval_stats_base import PredictionEvalStats, Metric, EvalStatsCollection, PredictionArray

log = logging.getLogger(__name__)


class RegressionMetric(Metric["RegressionEvalStats"], ABC):
    def computeValueForEvalStats(self, evalStats: "RegressionEvalStats"):
        return self.computeValue(np.array(evalStats.y_true), np.array(evalStats.y_predicted))

    @classmethod
    @abstractmethod
    def computeValue(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        pass

    @classmethod
    def computeErrors(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        return y_predicted - y_true

    @classmethod
    def computeAbsErrors(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        return np.abs(cls.computeErrors(y_true, y_predicted))


class RegressionMetricMAE(RegressionMetric):
    name = "MAE"

    @classmethod
    def computeValue(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        return np.mean(cls.computeAbsErrors(y_true, y_predicted))


class RegressionMetricMSE(RegressionMetric):
    name = "MSE"

    @classmethod
    def computeValue(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        residuals = y_predicted - y_true
        return np.sum(residuals * residuals) / len(residuals)


class RegressionMetricRMSE(RegressionMetric):
    name = "RMSE"

    @classmethod
    def computeValue(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        errors = cls.computeErrors(y_true, y_predicted)
        return np.sqrt(np.mean(errors * errors))


class RegressionMetricRRSE(RegressionMetric):
    name = "RRSE"

    @classmethod
    def computeValue(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        mean_y = np.mean(y_true)
        residuals = y_predicted - y_true
        mean_deviation = y_true - mean_y
        return np.sqrt(np.sum(residuals * residuals) / np.sum(mean_deviation * mean_deviation))


class RegressionMetricR2(RegressionMetric):
    name = "R2"

    def computeValue(self, y_true: np.ndarray, y_predicted: np.ndarray):
        rrse = RegressionMetricRRSE.computeValue(y_true, y_predicted)
        return 1.0 - rrse*rrse


class RegressionMetricPCC(RegressionMetric):
    name = "PCC"

    def computeValue(self, y_true: np.ndarray, y_predicted: np.ndarray):
        cov = np.cov([y_true, y_predicted])
        return cov[0][1] / np.sqrt(cov[0][0] * cov[1][1])


class RegressionMetricStdDevAE(RegressionMetric):
    name = "StdDevAE"

    @classmethod
    def computeValue(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        return np.std(cls.computeAbsErrors(y_true, y_predicted))


class RegressionMetricMedianAE(RegressionMetric):
    name = "MedianAE"

    @classmethod
    def computeValue(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        return np.median(cls.computeAbsErrors(y_true, y_predicted))


class RegressionEvalStats(PredictionEvalStats["RegressionMetric"]):
    """
    Collects data for the evaluation of predicted continuous values and computes corresponding metrics
    """

    # class members controlling plot appearance, which can be centrally overridden by a user if necessary
    HEATMAP_COLORMAP_FACTORY = lambda self: LinearSegmentedColormap.from_list("whiteToRed", ((0, (1, 1, 1)), (1/len(self.y_predicted), (1, 0.96, 0.96)), (1, (0.7, 0, 0))), len(self.y_predicted))
    HEATMAP_DIAGONAL_COLOR = "green"
    HEATMAP_ERROR_BOUNDARY_VALUE = None
    HEATMAP_ERROR_BOUNDARY_COLOR = (0.8, 0.8, 0.8)
    SCATTER_PLOT_POINT_COLOR = (0, 0, 1, 0.05)

    def __init__(self, y_predicted: Optional[PredictionArray] = None, y_true: Optional[PredictionArray] = None,
            metrics: Sequence["RegressionMetric"] = None, additionalMetrics: Sequence["RegressionMetric"] = None):
        """
        :param y_predicted: the predicted values
        :param y_true: the true values
        :param metrics: the metrics to compute for evaluation; if None, use default metrics
        :param additionalMetrics: the metrics to additionally compute
        """

        if metrics is None:
            metrics = [RegressionMetricRRSE(), RegressionMetricR2(), RegressionMetricPCC(),
                       RegressionMetricMAE(), RegressionMetricMSE(), RegressionMetricRMSE(),
                       RegressionMetricStdDevAE()]
        metrics = list(metrics)

        super().__init__(y_predicted, y_true, metrics, additionalMetrics=additionalMetrics)

    def getMSE(self):
        return self.computeMetricValue(RegressionMetricMSE())

    def getRRSE(self):
        """Gets the root relative squared error"""
        return self.computeMetricValue(RegressionMetricRRSE())

    def getCorrelationCoeff(self):
        """Gets the Pearson correlation coefficient (PCC)"""
        return self.computeMetricValue(RegressionMetricPCC())

    def getR2(self):
        """Gets the R^2 score"""
        return self.computeMetricValue(RegressionMetricR2())

    def getMAE(self):
        """Gets the mean absolute error"""
        return self.computeMetricValue(RegressionMetricMAE())

    def getRMSE(self):
        """Gets the root mean squared error"""
        return self.computeMetricValue(RegressionMetricRMSE())

    def getStdDevAE(self):
        """Gets the standard deviation of the absolute error"""
        return self.computeMetricValue(RegressionMetricStdDevAE())

    def getEvalStatsCollection(self) -> "RegressionEvalStatsCollection":
        """
        For the case where we collected data on multiple dimensions, obtain a stats collection where
        each object in the collection holds stats on just one dimension
        """
        if self.y_true_multidim is None:
            raise Exception("No multi-dimensional data was collected")
        dim = len(self.y_true_multidim)
        statsList = []
        for i in range(dim):
            stats = RegressionEvalStats(self.y_predicted_multidim[i], self.y_true_multidim[i])
            statsList.append(stats)
        return RegressionEvalStatsCollection(statsList)

    def plotErrorDistribution(self, bins=None, figure=True, titleAdd=None) -> Optional[plt.Figure]:
        """
        :param bins: if None, seaborns default binning will be used
        :param figure: whether to plot in a separate figure and return that figure
        :param titleAdd: a string to add to the title (on a second line)

        :return: the resulting figure object or None
        """
        errors = np.array(self.y_predicted) - np.array(self.y_true)
        fig = None
        title = "Prediction Error Distribution"
        if titleAdd is not None:
            title += "\n" + titleAdd
        if figure:
            fig = plt.figure(title.replace("\n", " "))
        sns.distplot(errors, bins=bins)
        plt.title(title)
        plt.xlabel("error (prediction - ground truth)")
        plt.ylabel("probability density")
        return fig

    def plotScatterGroundTruthPredictions(self, figure=True, titleAdd=None, **kwargs) -> Optional[plt.Figure]:
        """
        :param figure: whether to plot in a separate figure and return that figure
        :param titleAdd: a string to be added to the title in a second line
        :param kwargs: parameters to be passed on to plt.scatter()

        :return:  the resulting figure object or None
        """
        fig = None
        title = "Scatter Plot of Ground Truth vs. Predicted Values"
        if titleAdd is not None:
            title += "\n" + titleAdd
        if figure:
            fig = plt.figure(title.replace("\n", " "))
        y_range = [min(self.y_true), max(self.y_true)]
        plt.scatter(self.y_true, self.y_predicted, c=[self.SCATTER_PLOT_POINT_COLOR], zorder=2, **kwargs)
        plt.plot(y_range, y_range, '-', lw=1, label="_not in legend", color="green", zorder=1)
        plt.xlabel("ground truth")
        plt.ylabel("prediction")
        plt.title(title)
        return fig

    def plotHeatmapGroundTruthPredictions(self, figure=True, cmap=None, bins=60, titleAdd=None, errorBoundary: Optional[float] = None,
            **kwargs) -> Optional[plt.Figure]:
        """
        :param figure: whether to plot in a separate figure and return that figure
        :param cmap: the colour map to use (see corresponding parameter of plt.imshow for further information); if None, use factory
            defined in HEATMAP_COLORMAP_FACTORY (which can be centrally set to achieve custom behaviour throughout an application)
        :param bins: how many bins to use for constructing the heatmap
        :param titleAdd: a string to add to the title (on a second line)
        :param errorBoundary: if not None, add two lines (above and below the diagonal) indicating this absolute regression error boundary;
            if None (default), use static member HEATMAP_ERROR_BOUNDARY_VALUE (which is also None by default, but can be centrally set
            to achieve custom behaviour throughout an application)

        :param kwargs: will be passed to plt.imshow()

        :return:  the resulting figure object or None
        """
        fig = None
        title = "Heat Map of Ground Truth vs. Predicted Values"
        if titleAdd:
            title += "\n" + titleAdd
        if figure:
            fig = plt.figure(title.replace("\n", " "))
        y_range = [min(min(self.y_true), min(self.y_predicted)), max(max(self.y_true), max(self.y_predicted))]

        # diagonal
        plt.plot(y_range, y_range, '-', lw=0.75, label="_not in legend", color=self.HEATMAP_DIAGONAL_COLOR, zorder=2)

        # error boundaries
        if errorBoundary is None:
            errorBoundary = self.HEATMAP_ERROR_BOUNDARY_VALUE
        if errorBoundary is not None:
            d = np.array(y_range)
            offs = np.array([errorBoundary, errorBoundary])
            plt.plot(d, d + offs, '-', lw=0.75, label="_not in legend", color=self.HEATMAP_ERROR_BOUNDARY_COLOR, zorder=2)
            plt.plot(d, d - offs, '-', lw=0.75, label="_not in legend", color=self.HEATMAP_ERROR_BOUNDARY_COLOR, zorder=2)

        # heat map
        heatmap, _, _ = np.histogram2d(self.y_true, self.y_predicted, range=[y_range, y_range], bins=bins, density=False)
        extent = [y_range[0], y_range[1], y_range[0], y_range[1]]
        if cmap is None:
            cmap = self.HEATMAP_COLORMAP_FACTORY()
        plt.imshow(heatmap.T, extent=extent, origin='lower', interpolation="none", cmap=cmap, zorder=1, **kwargs)

        plt.xlabel("ground truth")
        plt.ylabel("prediction")
        plt.title(title)
        return fig


class RegressionEvalStatsCollection(EvalStatsCollection):
    def __init__(self, evalStatsList: List[RegressionEvalStats]):
        super().__init__(evalStatsList)
        self.globalStats = None

    def getGlobalStats(self) -> RegressionEvalStats:
        """
        Gets an evaluation statistics object that combines the data from all contained eval stats objects
        """
        if self.globalStats is None:
            y_true = np.concatenate([evalStats.y_true for evalStats in self.statsList])
            y_predicted = np.concatenate([evalStats.y_predicted for evalStats in self.statsList])
            self.globalStats = RegressionEvalStats(y_predicted, y_true)
        return self.globalStats



from enum import Enum
import logging
from typing import Callable, Union, TypeVar, Generic, Sequence, List, Tuple, Iterable, Dict, Hashable, Optional

import numpy as np

from .util.pickle import setstate
from .util.string import listString, ToStringMixin

T = TypeVar("T")

log = logging.getLogger(__name__)


class Vectoriser(Generic[T], ToStringMixin):
    """
    A vectoriser represents a method for the conversion of instances of some type T into
    vectors, i.e. one-dimensional (numeric) arrays, or (in the special case of a 1D vector) scalars
    """
    def __init__(self, f: Callable[[T], Union[float, np.ndarray, list]], transformer=None):
        """
        :param f: the function which maps from an instance of T to an array/list/scalar
        :param transformer: an optional transformer (e.g. instance of one of the classes in sklearn.preprocessing)
            which can be used to transform/normalise the generated arrays
        """
        self._fn = f
        self.transformer = transformer
        self._resultType = None
        self.name = None

    def __setstate__(self, state):
        setstate(Vectoriser, self, state, newOptionalProperties=["_resultType", "name"], renamedProperties={"f": "_fn"})

    def _toStringExcludePrivate(self) -> bool:
        return True

    def setName(self, name):
        self.name = name

    def getName(self):
        """
        :return: the name of this feature generator, which may be a default name if the name has not been set. Note that feature generators created
            by a FeatureGeneratorFactory always get the name with which the generator factory was registered.
        """
        if self.name is None:
            return f"{self.__class__.__name__}-{id(self)}"
        return self.name

    def fit(self, items: Iterable[T]):
        if self.transformer is not None:
            values = [self._f(item) for item in items]
            self.transformer.fit(np.array(values))

    def _f(self, x) -> np.array:
        y = self._fn(x)

        if self._resultType is None:
            self._resultType = self.ResultType.fromValue(y)

        if self._resultType == self.ResultType.LIST:
            y = np.array(y)
        elif self._resultType == self.ResultType.SCALAR:
            y = np.array([y])

        return y

    def apply(self, item: T, transform=True) -> np.array:
        """
        :param item: the item to be vectorised
        :param transform: whether to apply this instance's transformer (if any)
        :return: a vector
        """
        value = self._f(item)
        if self.transformer is not None and transform:
            value = self.transformer.transform([value])[0]
        return value

    def applyMulti(self, items: Iterable[T], transform=True) -> List[np.array]:
        """
        Applies this vectoriser to multiple items at once.
        Especially for cases where this vectoriser uses a transformer, this method is significantly faster than
        calling apply repeatedly.

        :param items: the items to be vectorised
        :param transform: whether to apply this instance's transformer (if any)
        :return: a list of vectors
        """
        values = [self._f(x) for x in items]
        if self.transformer is not None and transform:
            values = self.transformer.transform(values)
        return values

    class ResultType(Enum):
        SCALAR = 0
        LIST = 1
        NUMPY_ARRAY = 2

        @classmethod
        def fromValue(cls, y):
            if type(y) == list:
                return cls.LIST
            elif np.isscalar(y):
                return cls.SCALAR
            elif isinstance(y, np.ndarray):
                return cls.NUMPY_ARRAY
            else:
                raise ValueError(f"Received unhandled value of type {type(y)}")


class EmptyVectoriser(Vectoriser):
    def __init__(self):
        super().__init__(self._createEmptyVector)

    @staticmethod
    def _createEmptyVector(x):
        return np.zeros(0)


class SequenceVectoriser(Generic[T], ToStringMixin):
    """
    Supports the application of Vectorisers to sequences of objects of some type T, where each object of type T is
    mapped to a vector (1D array) by the vectorisers.
    A SequenceVectoriser is fitted by fitting the underlying Vectorisers. In order to obtain the instances of T that
    are used for training, we take into consideration the fact that the sequences of T may overlap and thus training
    is performed on the set of unique instances.
    """
    class FittingMode(Enum):
        """
        Determines how the individual vectorisers are fitted based on several sequences of objects of type T that are given.
        If NONE, no fitting is performed, otherwise the mode determines how a single sequence of objects of type T for fitting
        is obtained from the collection of sequences: either by forming the set of unique objects from the sequences (UNIQUE)
        """
        NONE = "none"  # no fitting is performed
        UNIQUE = "unique"  # use collection of unique items
        CONCAT = "concat"  # use collection obtained by concatenating all sequences using numpy.concatenate

    def __init__(self, vectorisers: Union[Sequence[Vectoriser[T]], Vectoriser[T]], fittingMode: FittingMode = FittingMode.UNIQUE):
        """
        :param vectorisers: zero or more vectorisers that are to be applied. If more than one vectoriser is supplied,
            vectors are generated from input instances of type T by concatenating the results of the vectorisers in
            the order the vectorisers are given.
        """
        self.fittingMode = fittingMode
        if isinstance(vectorisers, Vectoriser):
            self.vectorisers = [vectorisers]
        else:
            self.vectorisers = vectorisers
        if len(self.vectorisers) == 0:
            self.vectorisers = [EmptyVectoriser()]

    def __setstate__(self, state):
        state["fittingMode"] = state.get("fittingMode", self.FittingMode.UNIQUE)
        setstate(SequenceVectoriser, self, state)

    def fit(self, data: Iterable[Sequence[T]]):
        log.debug(f"Fitting {self}")
        if self.fittingMode == self.FittingMode.NONE:
            return
        if self.fittingMode == self.FittingMode.UNIQUE:
            items = set()
            for seq in data:
                items.update(seq)
        elif self.fittingMode == self.FittingMode.CONCAT:
            items = np.concatenate(data)
        else:
            raise ValueError(self.fittingMode)
        for v in self.vectorisers:
            log.debug(f"Fitting {v}")
            v.fit(items)

    def apply(self, seq: Sequence[T], transform=True) -> List[np.array]:
        """
        Applies vectorisation to the given sequence of objects

        :param seq: the sequence to vectorise
        :param transform: whether to apply any post-vectorisation transformers
        :return:
        """
        vectorsList = []
        for item in seq:
            vectors = [vec.apply(item, transform=transform) for vec in self.vectorisers]
            conc = np.concatenate(vectors, axis=0)
            vectorsList.append(conc)
        return vectorsList

    def applyMulti(self, sequences: Iterable[Sequence[T]]) -> Tuple[List[List[np.array]], List[int]]:
        """
        Applies this vectoriser to multiple sequences of objects of type T, where each sequence is mapped to a sequence
        of 1D arrays.
        This method can be significantly faster than multiple applications of apply, especially in cases where the vectorisers
        use transformers.

        :param sequences: the sequences to vectorise
        :return: a pair (vl, l) where vl is a list of lists of vectors/arrays and l is a list of integers containing the lengths
            of the sequences
        """
        lengths = [len(s) for s in sequences]

        combinedSeq = []
        for seq in sequences:
            combinedSeq.extend(seq)

        individualVectoriserResults = [vectoriser.applyMulti(combinedSeq) for vectoriser in self.vectorisers]
        concVectors = [np.concatenate(x, axis=0) for x in zip(*individualVectoriserResults)]

        vectorSequences = []
        idxStart = 0
        for l in lengths:
            vectorSequences.append(concVectors[idxStart:idxStart+l])
            idxStart += l

        return vectorSequences, lengths

    def applyMultiWithPadding(self, sequences: Sequence[Sequence[T]]) -> Tuple[List[List[np.array]], List[int]]:
        """
        Applies this vectoriser to multiple sequences of objects of type T, where each sequence is mapped to a sequence
        of 1D arrays.
        Sequences are allowed to vary in length. for shorter sequences, 0-vectors are appended until the maximum sequence length
        is reached (padding).

        :param sequences: the sequences to vectorise
        :return: a pair (vl, l) where vl is a list of lists of vectors/arrays, each list having the same length, and l is a list of
            integers containing the original unpadded lengths of the sequences
        """
        result, lengths = self.applyMulti(sequences)
        maxLength = max(lengths)
        dim = len(result[0][0])
        dummyVec = np.zeros((dim,))
        for seq in result:
            for i in range(maxLength - len(seq)):
                seq.append(dummyVec)
        return result, lengths

    def getVectorDim(self, seq: Sequence[T]):
        """
        Determines the dimensionality of generated vectors by applying the vectoriser to the given sequence

        :param seq: the sequence
        :return: the number of dimensions in generated output vectors (per item)
        """
        return len(self.apply(seq, transform=False)[0])


if __name__ == '__main__':
    def myf(x):
        return np.array([x/2, x*x])

    import sklearn.preprocessing

    items = [1,2,3]
    items2 = [4,5,6,7]
    data = [items, items2]
    vectoriser = Vectoriser(myf, transformer=sklearn.preprocessing.MaxAbsScaler())
    #vectoriser.fit(items)
    #result = vectoriser.apply(items[1])

    svec = SequenceVectoriser([vectoriser, vectoriser])
    svec.fit(data)
    result, lengths = svec.applyMultiWithPadding(data)


class VectoriserRegistry:
    def __init__(self):
        self._factories: Dict[Hashable, Callable[[Callable], Vectoriser]] = {}

    def getAvailableVectorisers(self):
        return list(self._factories.keys())

    @staticmethod
    def _name(name: Hashable):
        # for enums, which have .name, use the name only, because it is less problematic to persist
        if hasattr(name, "name"):
            name = name.name
        return name

    def registerFactory(self, name: Hashable, factory: Callable[[Callable], Vectoriser],
            additionalNames: Optional[Iterable[Hashable]] = None):
        """
        Registers a vectoriser factory which can subsequently be referenced via their name

        :param name: the name (which can, in particular, be a string or an enum item)
        :param factory: the factory, which takes the default transformer factory as an argument
        :param additionalNames: (optional) additional names under which to register the factory
        """
        self._registerFactory(name, factory)
        if additionalNames is not None:
            for n in additionalNames:
                self._registerFactory(n, factory)

    def _registerFactory(self, name: Hashable, factory):
        name = self._name(name)
        if name in self._factories:
            raise ValueError(f"Vectoriser factory for name '{name}' already registered")
        self._factories[name] = factory

    def getVectoriser(self, name: Hashable, defaultTransformerFactory: Callable) -> Vectoriser:
        """
        Creates a vectoriser from a name, which must have been previously registered.

        :param name: the name (which can, in particular, be a string or an enum item)
        :param defaultTransformerFactory: the default transformer factory
        :return: a new vectoriser instance
        """
        name = self._name(name)
        factory = self._factories.get(name)
        if factory is None:
            raise ValueError(f"No vectoriser factory registered for name '{name}': known names: {listString(self._factories.keys())}. Register the factory first.")
        instance = factory(defaultTransformerFactory)
        instance.setName(name)
        return instance

    def getVectorisers(self, names: List[Hashable], defaultTransformerFactory: Callable) -> List[Vectoriser]:
        return [self.getVectoriser(name, defaultTransformerFactory) for name in names]


import copy
import logging
import re
from abc import ABC, abstractmethod
from typing import List, Sequence, Union, Dict, Callable, Any, Optional, Set

import numpy as np
import pandas as pd
from sklearn.preprocessing import OneHotEncoder

from .sklearn_transformer import SkLearnTransformerProtocol
from ..columngen import ColumnGenerator
from ..util import flattenArguments
from ..util.pandas import DataFrameColumnChangeTracker
from ..util.pickle import setstate
from ..util.string import orRegexGroup, ToStringMixin

log = logging.getLogger(__name__)


class DataFrameTransformer(ABC, ToStringMixin):
    """
    Base class for data frame transformers, i.e. objects which can transform one data frame into another
    (possibly applying the transformation to the original data frame - in-place transformation).
    A data frame transformer may require being fitted using training data.
    """
    def __init__(self):
        self._name = f"{self.__class__.__name__}-{id(self)}"
        self._isFitted = False
        self._columnChangeTracker: Optional[DataFrameColumnChangeTracker] = None
        self._paramInfo = {}  # arguments passed to init that are not saved otherwise can be persisted here

    # for backwards compatibility with persisted DFTs based on code prior to commit 7088cbbe
    # They lack the __isFitted attribute and we assume that each such DFT was fitted
    def __setstate__(self, d):
        d["_name"] = d.get("_name", f"{self.__class__.__name__}-{id(self)}")
        d["_isFitted"] = d.get("_isFitted", True)
        d["_columnChangeTracker"] = d.get("_columnChangeTracker", None)
        d["_paramInfo"] = d.get("_paramInfo", {})
        self.__dict__ = d

    def _toStringExcludePrivate(self) -> bool:
        return True

    def getName(self) -> str:
        """
        :return: the name of this dft transformer, which may be a default name if the name has not been set.
        """
        return self._name

    def setName(self, name):
        self._name = name

    @abstractmethod
    def _fit(self, df: pd.DataFrame):
        pass

    @abstractmethod
    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    def apply(self, df: pd.DataFrame) -> pd.DataFrame:
        self._columnChangeTracker = DataFrameColumnChangeTracker(df)
        if not self.isFitted():
            raise Exception(f"Cannot apply a DataFrameTransformer which is not fitted: "
                            f"the df transformer {self.getName()} requires fitting")
        df = self._apply(df)
        self._columnChangeTracker.trackChange(df)
        return df

    def info(self):
        return {
            "name": self.getName(),
            "changeInColumnNames": self._columnChangeTracker.columnChangeString() if self._columnChangeTracker is not None else None,
            "isFitted": self.isFitted(),
        }

    def fit(self, df: pd.DataFrame):
        self._fit(df)
        self._isFitted = True

    def isFitted(self):
        return self._isFitted

    def fitApply(self, df: pd.DataFrame) -> pd.DataFrame:
        self.fit(df)
        return self.apply(df)


class InvertibleDataFrameTransformer(DataFrameTransformer, ABC):
    @abstractmethod
    def applyInverse(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    def getInverse(self) -> "InverseDataFrameTransformer":
        """
        :return: a transformer whose (forward) transformation is the inverse transformation of this DFT
        """
        return InverseDataFrameTransformer(self)


class RuleBasedDataFrameTransformer(DataFrameTransformer, ABC):
    """Base class for transformers whose logic is entirely based on rules and does not need to be fitted to data"""

    def _fit(self, df: pd.DataFrame):
        pass

    def fit(self, df: pd.DataFrame):
        pass

    def isFitted(self):
        return True


class InverseDataFrameTransformer(RuleBasedDataFrameTransformer):
    def __init__(self, invertibleDFT: InvertibleDataFrameTransformer):
        super().__init__()
        self.invertibleDFT = invertibleDFT

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return self.invertibleDFT.applyInverse(df)


class DataFrameTransformerChain(DataFrameTransformer):
    """
    Supports the application of a chain of data frame transformers.
    During fit and apply each transformer in the chain receives the transformed output of its predecessor.
    """

    def __init__(self, *dataFrameTransformers: Union[DataFrameTransformer, List[DataFrameTransformer]]):
        super().__init__()
        self.dataFrameTransformers = flattenArguments(dataFrameTransformers)

    def __len__(self):
        return len(self.dataFrameTransformers)

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        for transformer in self.dataFrameTransformers:
            df = transformer.apply(df)
        return df

    def _fit(self, df: pd.DataFrame):
        if len(self.dataFrameTransformers) == 0:
            return
        for transformer in self.dataFrameTransformers[:-1]:
            df = transformer.fitApply(df)
        self.dataFrameTransformers[-1].fit(df)

    def isFitted(self):
        return all([dft.isFitted() for dft in self.dataFrameTransformers])

    def getNames(self) -> List[str]:
        """
        :return: the list of names of all contained feature generators
        """
        return [transf.getName() for transf in self.dataFrameTransformers]

    def info(self):
        info = super().info()
        info["chainedDFTTransformerNames"] = self.getNames()
        info["length"] = len(self)
        return info

    def findFirstTransformerByType(self, cls) -> Optional[DataFrameTransformer]:
        for dft in self.dataFrameTransformers:
            if isinstance(dft, cls):
                return dft
        return None


class DFTRenameColumns(RuleBasedDataFrameTransformer):
    def __init__(self, columnsMap: Dict[str, str]):
        """
        :param columnsMap: dictionary mapping old column names to new names
        """
        super().__init__()
        self.columnsMap = columnsMap

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.rename(columns=self.columnsMap)


class DFTConditionalRowFilterOnColumn(RuleBasedDataFrameTransformer):
    """
    Filters a data frame by applying a boolean function to one of the columns and retaining only the rows
    for which the function returns True
    """
    def __init__(self, column: str, condition: Callable[[Any], bool]):
        super().__init__()
        self.column = column
        self.condition = condition

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df[df[self.column].apply(self.condition)]


class DFTInSetComparisonRowFilterOnColumn(RuleBasedDataFrameTransformer):
    """
    Filters a data frame on the selected column and retains only the rows for which the value is in the setToKeep
    """
    def __init__(self, column: str, setToKeep: Set):
        super().__init__()
        self.setToKeep = setToKeep
        self.column = column

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df[df[self.column].isin(self.setToKeep)]

    def info(self):
        info = super().info()
        info["column"] = self.column
        info["setToKeep"] = self.setToKeep
        return info


class DFTNotInSetComparisonRowFilterOnColumn(RuleBasedDataFrameTransformer):
    """
    Filters a data frame on the selected column and retains only the rows for which the value is not in the setToDrop
    """
    def __init__(self, column: str, setToDrop: Set):
        super().__init__()
        self.setToDrop = setToDrop
        self.column = column

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df[~df[self.column].isin(self.setToDrop)]

    def info(self):
        info = super().info()
        info["column"] = self.column
        info["setToDrop"] = self.setToDrop
        return info


class DFTVectorizedConditionalRowFilterOnColumn(RuleBasedDataFrameTransformer):
    """
    Filters a data frame by applying a vectorized condition on the selected column and retaining only the rows
    for which it returns True
    """
    def __init__(self, column: str, vectorizedCondition: Callable[[pd.Series], Sequence[bool]]):
        super().__init__()
        self.column = column
        self.vectorizedCondition = vectorizedCondition

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df[self.vectorizedCondition(df[self.column])]

    def info(self):
        info = super().info()
        info["column"] = self.column
        return info


class DFTRowFilter(RuleBasedDataFrameTransformer):
    """
    Filters a data frame by applying a condition function to each row and retaining only the rows
    for which it returns True
    """
    def __init__(self, condition: Callable[[Any], bool]):
        super().__init__()
        self.condition = condition

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df[df.apply(self.condition, axis=1)]


class DFTModifyColumn(RuleBasedDataFrameTransformer):
    """
    Modifies a column specified by 'column' using 'columnTransform'
    """
    def __init__(self, column: str, columnTransform: Union[Callable, np.ufunc]):
        """
        :param column: the name of the column to be modified
        :param columnTransform: a function operating on single cells or a Numpy ufunc that applies to an entire Series
        """
        super().__init__()
        self.column = column
        self.columnTransform = columnTransform

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df[self.column] = df[self.column].apply(self.columnTransform)
        return df


class DFTModifyColumnVectorized(RuleBasedDataFrameTransformer):
    """
    Modifies a column specified by 'column' using 'columnTransform'. This transformer can be used to utilise Numpy vectorisation for
    performance optimisation.
    """
    def __init__(self, column: str, columnTransform: Callable[[np.ndarray], Union[Sequence, pd.Series, np.ndarray]]):
        """
        :param column: the name of the column to be modified
        :param columnTransform: a function that takes a Numpy array and from which the returned value will be assigned to the column as a whole
        """
        super().__init__()
        self.column = column
        self.columnTransform = columnTransform

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df[self.column] = self.columnTransform(df[self.column].values)
        return df


class DFTOneHotEncoder(DataFrameTransformer):
    def __init__(self, columns: Optional[Union[str, Sequence[str]]],
            categories: Union[List[np.ndarray], Dict[str, np.ndarray]] = None, inplace=False, ignoreUnknown=False,
            arrayValuedResult=False):
        """
        One hot encode categorical variables

        :param columns: list of names or regex matching names of columns that are to be replaced by a list one-hot encoded columns each
            (or an array-valued column for the case where useArrayValues=True);
            If None, then no columns are actually to be one-hot-encoded
        :param categories: numpy arrays containing the possible values of each of the specified columns (for case where sequence is specified
            in 'columns') or dictionary mapping column name to array of possible categories for the column name.
            If None, the possible values will be inferred from the columns
        :param inplace: whether to perform the transformation in-place
        :param ignoreUnknown: if True and an unknown category is encountered during transform, the resulting one-hot
            encoded columns for this feature will be all zeros. if False, an unknown category will raise an error.
        :param arrayValuedResult: whether to replace the input columns by columns of the same name containing arrays as values
            instead of creating a separate column per original value
        """
        super().__init__()
        self._paramInfo["columns"] = columns
        self._paramInfo["inferCategories"] = categories is None
        self.oneHotEncoders = None
        if columns is None:
            self._columnsToEncode = []
            self._columnNameRegex = "$"
        elif type(columns) == str:
            self._columnNameRegex = columns
            self._columnsToEncode = None
        else:
            self._columnNameRegex = orRegexGroup(columns)
            self._columnsToEncode = columns
        self.inplace = inplace
        self.arrayValuedResult = arrayValuedResult
        self.handleUnknown = "ignore" if ignoreUnknown else "error"
        if categories is not None:
            if type(categories) == dict:
                self.oneHotEncoders = {col: OneHotEncoder(categories=[np.sort(categories)], sparse=False, handle_unknown=self.handleUnknown) for col, categories in categories.items()}
            else:
                if len(columns) != len(categories):
                    raise ValueError(f"Given categories must have the same length as columns to process")
                self.oneHotEncoders = {col: OneHotEncoder(categories=[np.sort(categories)], sparse=False, handle_unknown=self.handleUnknown) for col, categories in zip(columns, categories)}

    def __setstate__(self, state):
        if "arrayValuedResult" not in state:
            state["arrayValuedResult"] = False
        super().__setstate__(state)

    def _toStringAdditionalEntries(self) -> Dict[str, Any]:
        d = super()._toStringAdditionalEntries()
        d["columns"] = self._paramInfo.get("columns")
        return d

    def _fit(self, df: pd.DataFrame):
        if self._columnsToEncode is None:
            self._columnsToEncode = [c for c in df.columns if re.fullmatch(self._columnNameRegex, c) is not None]
            if len(self._columnsToEncode) == 0:
                log.warning(f"{self} does not apply to any columns, transformer has no effect; regex='{self._columnNameRegex}'")
        if self.oneHotEncoders is None:
            self.oneHotEncoders = {column: OneHotEncoder(categories=[np.sort(df[column].unique())], sparse=False, handle_unknown=self.handleUnknown) for column in self._columnsToEncode}
        for columnName in self._columnsToEncode:
            self.oneHotEncoders[columnName].fit(df[[columnName]])

    def _apply(self, df: pd.DataFrame):
        if len(self._columnsToEncode) == 0:
            return df

        if not self.inplace:
            df = df.copy()
        for columnName in self._columnsToEncode:
            encodedArray = self.oneHotEncoders[columnName].transform(df[[columnName]])
            if not self.arrayValuedResult:
                df = df.drop(columns=columnName)
                for i in range(encodedArray.shape[1]):
                    df["%s_%d" % (columnName, i)] = encodedArray[:, i]
            else:
                df[columnName] = list(encodedArray)
        return df

    def info(self):
        info = super().info()
        info["inplace"] = self.inplace
        info["handleUnknown"] = self.handleUnknown
        info["arrayValuedResult"] = self.arrayValuedResult
        info.update(self._paramInfo)
        return info


class DFTColumnFilter(RuleBasedDataFrameTransformer):
    """
    A DataFrame transformer that filters columns by retaining or dropping specified columns
    """
    def __init__(self, keep: Union[str, Sequence[str]] = None, drop: Union[str, Sequence[str]] = None):
        super().__init__()
        self.keep = [keep] if type(keep) == str else keep
        self.drop = drop

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        if self.keep is not None:
            df = df[self.keep]
        if self.drop is not None:
            df = df.drop(columns=self.drop)
        return df

    def info(self):
        info = super().info()
        info["keep"] = self.keep
        info["drop"] = self.drop
        return info


class DFTKeepColumns(DFTColumnFilter):
    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df[self.keep]


class DFTDRowFilterOnIndex(RuleBasedDataFrameTransformer):
    def __init__(self, keep: Set = None, drop: Set = None):
        super().__init__()
        self.drop = drop
        self.keep = keep

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        if self.keep is not None:
            df = df.loc[self.keep]
        if self.drop is not None:
            df = df.drop(self.drop)
        return df


class DFTNormalisation(DataFrameTransformer):
    """
    Applies normalisation/scaling to a data frame by applying a set of transformation rules, where each
    rule defines a set of columns to which it applies (learning a single transformer based on the values
    of all applicable columns)
    """

    class RuleTemplate:
        def __init__(self, skip=False, unsupported=False, transformer: SkLearnTransformerProtocol = None,
                transformerFactory: Callable[[], SkLearnTransformerProtocol] = None, independentColumns=False):
            """
            :param skip: flag indicating whether no transformation shall be performed on all of the columns
            :param unsupported: flag indicating whether normalisation of all columns is unsupported (shall trigger an exception if attempted)
            :param transformer: a transformer instance (from sklearn.preprocessing, e.g. StandardScaler) to apply to the matching column(s)
                for the case where a transformation is necessary (skip=False, unsupported=False). If None is given, either transformerFactory
                or the containing instance's default factory will be used.
                NOTE: Use an instance only if you want, in particular, the instance to be shared across several models that use the same
                feature with associated rule/rule template (disabling `fit` where appropriate). Otherwise, use a factory.
            :param transformerFactory: a factory for the generation of the transformer instance, which will only be applied if `transformer`
                is not given; if neither `transformer` nor `transformerInstance` are given, the containing instance's default factory will
                be used. See `SkLearnTransformerFactoryFactory` for convenient construction options.
            :param independentColumns: whether a separate transformation is to be learned for each of the columns for the case where the
                rule matches multiple columns.
            """
            if skip and transformer is not None:
                raise ValueError("skip==True while transformer is not None")
            self.skip = skip
            self.unsupported = unsupported
            self.transformer = transformer
            self.transformerFactory = transformerFactory
            self.independentColumns = independentColumns

        def toRule(self, regex: Optional[str]):
            """
            Convert the template to a rule for all columns matching the regex

            :param regex: a regular expression defining the column the rule applies to
            :return: the resulting Rule
            """
            return DFTNormalisation.Rule(regex, skip=self.skip, unsupported=self.unsupported, transformer=self.transformer,
                transformerFactory=self.transformerFactory, independentColumns=self.independentColumns)

        def toPlaceholderRule(self):
            return self.toRule(None)

    class Rule(ToStringMixin):
        def __init__(self, regex: Optional[str], skip=False, unsupported=False, transformer: SkLearnTransformerProtocol = None,
                transformerFactory: Callable[[], SkLearnTransformerProtocol] = None,
                arrayValued=False, fit=True, independentColumns=False):
            """
            :param regex: a regular expression defining the column(s) the rule applies to.
                If it applies to multiple columns, these columns will be normalised in the same way (using the same normalisation
                process for each column) unless independentColumns=True.
                If None, the rule is a placeholder rule and the regex must be set later via setRegex or the rule will not be applicable.
            :param skip: flag indicating whether no transformation shall be performed on the matching column(s)
            :param unsupported: flag indicating whether normalisation of the matching column(s) is unsupported (shall trigger an exception if attempted)
            :param transformer: a transformer instance (from sklearn.preprocessing, e.g. StandardScaler) to apply to the matching column(s)
                for the case where a transformation is necessary (skip=False, unsupported=False). If None is given, either transformerFactory
                or the containing instance's default factory will be used.
                NOTE: Use an instance only if you want, in particular, the instance to be shared across several models that use the same
                feature with associated rule/rule template (disabling `fit` where appropriate). Otherwise, use a factory.
            :param transformerFactory: a factory for the generation of the transformer instance, which will only be applied if `transformer`
                is not given; if neither `transformer` nor `transformerInstance` are given, the containing instance's default factory will
                be used. See `SkLearnTransformerFactoryFactory` for convenient construction options.
            :param arrayValued: whether the column values are not scalars but arrays (of arbitrary lengths).
                It is assumed that all entries in such arrays are to be normalised in the same way.
                If arrayValued is True, only a single matching column is supported, i.e. the regex must match at most one column.
            :param fit: whether the rule's transformer shall be fitted
            :param independentColumns: whether a separate transformation is to be learned for each of the columns for the case where the
                rule matches multiple columns.
            """
            if skip and (transformer is not None or transformerFactory is not None):
                raise ValueError("skip==True while transformer/transformerFactory is not None")
            self.regex = re.compile(regex) if regex is not None else None
            self.skip = skip
            self.unsupported = unsupported
            self.transformer = transformer
            self.transformerFactory = transformerFactory
            self.arrayValued = arrayValued
            self.fit = fit
            self.independentColumns = independentColumns

        def __setstate__(self, state):
            setstate(DFTNormalisation.Rule, self, state, newDefaultProperties=dict(arrayValued=False, fit=True, independentColumns=False,
                    transformerFactory=None))

        def _toStringExcludes(self) -> List[str]:
            return super()._toStringExcludes() + ["regex"]

        def _toStringAdditionalEntries(self) -> Dict[str, Any]:
            d = super()._toStringAdditionalEntries()
            if self.regex is not None:
                d["regex"] = f"'{self.regex.pattern}'"
            return d

        def setRegex(self, regex: str):
            self.regex = re.compile(regex)

        def matches(self, column: str):
            if self.regex is None:
                raise Exception("Attempted to apply a placeholder rule. Perhaps the feature generator from which the rule originated was never applied in order to have the rule instantiated.")
            return self.regex.fullmatch(column) is not None

        def matchingColumns(self, columns: Sequence[str]):
            return [col for col in columns if self.matches(col)]

    def __init__(self, rules: Sequence[Rule], defaultTransformerFactory=None, requireAllHandled=True, inplace=False):
        """
        :param rules: the set of rules; rules are always fitted and applied in the given order
        :param defaultTransformerFactory: a factory for the creation of transformer instances (from sklearn.preprocessing, e.g. StandardScaler)
            that shall be used to create a transformer for all rules that don't specify a particular transformer.
            The default transformer will only be applied to columns matched by such rules, unmatched columns will
            not be transformed.
        :param requireAllHandled: whether to raise an exception if not all columns are matched by a rule
        :param inplace: whether to apply data frame transformations in-place
        """
        super().__init__()
        self.requireAllHandled = requireAllHandled
        self.inplace = inplace
        self._userRules = rules
        self._defaultTransformerFactory = defaultTransformerFactory
        self._rules = None

    def _toStringAdditionalEntries(self) -> Dict[str, Any]:
        d = super()._toStringAdditionalEntries()
        if self._rules is not None:
            d["rules"] = self._rules
        else:
            d["userRules"] = self._userRules
        return d

    def _fit(self, df: pd.DataFrame):
        matchedRulesByColumn = {}
        self._rules = []
        for rule in self._userRules:
            matchingColumns = rule.matchingColumns(df.columns)
            for c in matchingColumns:
                if c in matchedRulesByColumn:
                    raise Exception(f"More than one rule applies to column '{c}': {matchedRulesByColumn[c]}, {rule}")
                matchedRulesByColumn[c] = rule

            if len(matchingColumns) > 0:
                if rule.unsupported:
                    raise Exception(f"Normalisation of columns {matchingColumns} is unsupported according to {rule}. If you want to make use of these columns, transform them into a supported column before applying {self.__class__.__name__}.")
                if not rule.skip:
                    if rule.transformer is None:
                        if rule.transformerFactory is not None:
                            rule.transformer = rule.transformerFactory()
                        else:
                            if self._defaultTransformerFactory is None:
                                raise Exception(f"No transformer to fit: {rule} defines no transformer and instance has no transformer factory")
                            rule.transformer = self._defaultTransformerFactory()
                    if rule.fit:
                        # fit transformer
                        applicableDF = df[sorted(matchingColumns)]
                        if rule.arrayValued:
                            if len(matchingColumns) > 1:
                                raise Exception(f"Array-valued case is only supported for a single column, matched {matchingColumns} for {rule}")
                            values = np.concatenate(applicableDF.values.flatten())
                            values = values.reshape((len(values), 1))
                        elif rule.independentColumns:
                            values = applicableDF.values
                        else:
                            values = applicableDF.values.flatten()
                            values = values.reshape((len(values), 1))
                        rule.transformer.fit(values)
            else:
                log.log(logging.DEBUG - 1, f"{rule} matched no columns")

            # collect specialised rule for application
            specialisedRule = copy.copy(rule)
            r = orRegexGroup(matchingColumns)
            try:
                specialisedRule.regex = re.compile(r)
            except Exception as e:
                raise Exception(f"Could not compile regex '{r}': {e}")
            self._rules.append(specialisedRule)

    def _checkUnhandledColumns(self, df, matchedRulesByColumn):
        if self.requireAllHandled:
            unhandledColumns = set(df.columns) - set(matchedRulesByColumn.keys())
            if len(unhandledColumns) > 0:
                raise Exception(f"The following columns are not handled by any rules: {unhandledColumns}; rules: {', '.join(map(str, self._rules))}")

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        if not self.inplace:
            df = df.copy()
        matchedRulesByColumn = {}
        for rule in self._rules:
            matchingColumns = rule.matchingColumns(df.columns)
            if len(matchingColumns) == 0:
                continue
            for c in matchingColumns:
                matchedRulesByColumn[c] = rule
            if not rule.skip:
                if rule.independentColumns and not rule.arrayValued:
                    matchingColumns = sorted(matchingColumns)
                    df[matchingColumns] = rule.transformer.transform(df[matchingColumns].values)
                else:
                    for c in matchingColumns:
                        if not rule.arrayValued:
                            df[c] = rule.transformer.transform(df[[c]].values)
                        else:
                            df[c] = [rule.transformer.transform(np.array([x]).T)[:, 0] for x in df[c]]
        self._checkUnhandledColumns(df, matchedRulesByColumn)
        return df

    def info(self):
        info = super().info()
        info["requireAllHandled"] = self.requireAllHandled
        info["inplace"] = self.inplace
        return info

    def findRule(self, colName: str) -> "DFTNormalisation.Rule":
        for rule in self._rules:
            if rule.matches(colName):
                return rule


class DFTFromColumnGenerators(RuleBasedDataFrameTransformer):
    def __init__(self, columnGenerators: Sequence[ColumnGenerator], inplace=False):
        super().__init__()
        self.columnGenerators = columnGenerators
        self.inplace = inplace

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        if not self.inplace:
            df = df.copy()
        for cg in self.columnGenerators:
            series = cg.generateColumn(df)
            df[series.name] = series
        return df

    def info(self):
        info = super().info()
        info["inplace"] = self.inplace
        return info


class DFTCountEntries(RuleBasedDataFrameTransformer):
    """
    Adds a new column with counts of the values on a selected column
    """
    def __init__(self, columnForEntryCount: str, columnNameForResultingCounts: str = "counts"):
        super().__init__()
        self.columnNameForResultingCounts = columnNameForResultingCounts
        self.columnForEntryCount = columnForEntryCount

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        series = df[self.columnForEntryCount].value_counts()
        return pd.DataFrame({self.columnForEntryCount: series.index, self.columnNameForResultingCounts: series.values})

    def info(self):
        info = super().info()
        info["columnNameForResultingCounts"] = self.columnNameForResultingCounts
        info["columnForEntryCount"] = self.columnForEntryCount
        return info


class DFTAggregationOnColumn(RuleBasedDataFrameTransformer):
    def __init__(self, columnForAggregation: str, aggregation: Callable):
        super().__init__()
        self.columnForAggregation = columnForAggregation
        self.aggregation = aggregation

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.groupby(self.columnForAggregation).agg(self.aggregation)


class DFTRoundFloats(RuleBasedDataFrameTransformer):
    def __init__(self, decimals=0):
        super().__init__()
        self.decimals = decimals

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame(np.round(df.values, self.decimals), columns=df.columns, index=df.index)

    def info(self):
        info = super().info()
        info["decimals"] = self.decimals
        return info


class DFTSkLearnTransformer(InvertibleDataFrameTransformer):
    """
    Applies a transformer from sklearn.preprocessing to (a subset of the columns of) a data frame.
    If multiple columns are transformed, they are transformed independently (i.e. each column uses a separately trained transformation).
    """
    def __init__(self, sklearnTransformer: SkLearnTransformerProtocol, columns: Optional[List[str]] = None, inplace=False,
            arrayValued=False):
        """
        :param sklearnTransformer: the transformer instance (from sklearn.preprocessing) to use (which will be fitted & applied)
        :param columns: the set of column names to which the transformation shall apply; if None, apply it to all columns
        :param inplace: whether to apply the transformation in-place
        :param arrayValued: whether to apply transformation not to scalar-valued columns but to one or more array-valued columns,
            where the values of all arrays within a column (which may vary in length) are to be transformed in the same way.
            If multiple columns are transformed, then the arrays belonging to a single row must all have the same length.
        """
        super().__init__()
        self.setName(f"{self.__class__.__name__}_wrapped_{sklearnTransformer.__class__.__name__}")
        self.sklearnTransformer = sklearnTransformer
        self.columns = columns
        self.inplace = inplace
        self.arrayValued = arrayValued

    def __setstate__(self, state):
        state["arrayValued"] = state.get("arrayValued", False)
        setstate(DFTSkLearnTransformer, self, state)

    def _fit(self, df: pd.DataFrame):
        cols = self.columns
        if cols is None:
            cols = df.columns
        if not self.arrayValued:
            values = df[cols].values
        else:
            if len(cols) == 1:
                values = np.concatenate(df[cols[0]].values.flatten())
                values = values.reshape((len(values), 1))
            else:
                flatColArrays = [np.concatenate(df[col].values.flatten()) for col in cols]
                lengths = [len(a) for a in flatColArrays]
                if len(set(lengths)) != 1:
                    raise ValueError(f"Columns {cols} do not contain the same number of values: {lengths}")
                values = np.stack(flatColArrays, axis=1)
        self.sklearnTransformer.fit(values)

    def _apply_transformer(self, df: pd.DataFrame, inverse: bool) -> pd.DataFrame:
        if not self.inplace:
            df = df.copy()
        cols = self.columns
        if cols is None:
            cols = df.columns
        transform = (lambda x: self.sklearnTransformer.inverse_transform(x)) if inverse else lambda x: self.sklearnTransformer.transform(x)
        if not self.arrayValued:
            df[cols] = transform(df[cols].values)
        else:
            if len(cols) == 1:
                c = cols[0]
                df[c] = [transform(np.array([x]).T)[:, 0] for x in df[c]]
            else:
                transformedValues = [transform(np.stack(row, axis=1)) for row in df.values]
                for iCol, col in enumerate(cols):
                    df[col] = [row[:, iCol] for row in transformedValues]
        return df

    def _apply(self, df):
        return self._apply_transformer(df, False)

    def applyInverse(self, df):
        return self._apply_transformer(df, True)

    def info(self):
        info = super().info()
        info["columns"] = self.columns
        info["inplace"] = self.inplace
        info["sklearnTransformerClass"] = self.sklearnTransformer.__class__.__name__
        return info


class DFTSortColumns(RuleBasedDataFrameTransformer):
    """
    Sorts a data frame's columns in ascending order
    """
    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df[sorted(df.columns)]



import collections
from math import log, exp

import numpy as np
import pandas as pd

from .vector_model import VectorClassificationModel


class CategoricalNaiveBayesVectorClassificationModel(VectorClassificationModel):
    """
    Naive Bayes with categorical features
    """
    def __init__(self, pseudoCount=0.1):
        """
        :param pseudoCount: the count to add to each empirical count in order to avoid overfitting
        """
        super().__init__()
        self.prior = None
        self.conditionals = None
        self.pseudoCount = pseudoCount

    def _fitClassifier(self, X: pd.DataFrame, y: pd.DataFrame):
        self.prior = collections.defaultdict(lambda: 0)
        self.conditionals = collections.defaultdict(lambda: [collections.defaultdict(lambda: 0) for _ in range(X.shape[1])])
        increment = 1
        for idxRow in range(X.shape[0]):
            cls = y.iloc[idxRow,0]
            self.prior[cls] += increment
            for idxFeature in range(X.shape[1]):
                value = X.iloc[idxRow, idxFeature]
                self.conditionals[cls][idxFeature][value] += increment

    def _predictClassProbabilities(self, X: pd.DataFrame):
        results = []
        for _, features in X.iterrows():
            classProbabilities = np.zeros(len(self._labels))
            for i, cls in enumerate(self._labels):
                lp = log(self._probability(self.prior, cls))
                for idxFeature, value in enumerate(features):
                    lp += log(self._probability(self.conditionals[cls][idxFeature], value))
                classProbabilities[i] = exp(lp)
            classProbabilities /= np.sum(classProbabilities)
            results.append(classProbabilities)
        return pd.DataFrame(results, columns=self._labels)

    def _probability(self, counts, value):
        valueCount = counts.get(value, 0.0)
        totalCount = sum(counts.values())
        return (valueCount + self.pseudoCount) / (totalCount + self.pseudoCount)

    def _predict(self, X: pd.DataFrame) -> pd.DataFrame:
        results = []
        for _, features in X.iterrows():
            bestCls = None
            bestLp = None
            for cls in self.prior:
                lp = log(self._probability(self.prior, cls))
                for idxFeature, value in enumerate(features):
                    lp += log(self._probability(self.conditionals[cls][idxFeature], value))
                if bestLp is None or lp > bestLp:
                    bestLp = lp
                    bestCls = cls
            results.append(bestCls)
        return pd.DataFrame(results, columns=self.getPredictedVariableNames())



"""
This module contains utilities for retrieving and visualizing ground truth labels for evaluating clustering algorithms
"""

import geopandas as gp
import logging
import numpy as np
from geopandas import GeoDataFrame
from shapely.geometry import Polygon, MultiPoint, MultiPolygon
from typing import Sequence, Union, Optional

from .coordinates import extractCoordinatesArray, TCoordinates, GeoDataFrameWrapper

log = logging.getLogger(__name__)


class PolygonAnnotatedCoordinates(GeoDataFrameWrapper):
    """
    Class for retrieving ground truth cluster labels from a set of coordinate points and polygons.
    From the provided 2-dim. coordinates only points within the ground truth region will be considered.
    """

    def __init__(self, coordinates: TCoordinates, groundTruthPolygons: Union[str, Sequence[Polygon], GeoDataFrame],
                 noiseLabel: Optional[int] = -1):
        """
        :param coordinates: coordinates of points. These points should be spread over an area larger or equal to
            the ground truth area
        :param groundTruthPolygons: sequence of polygons, GeoDataFrame or path to a shapefile containing such a sequence.
            The polygons represent the ground truth for clustering.
            *Important*: the first polygon in the sequence is assumed to be the region within
            which ground truth was provided and has to cover all remaining polygons. This also means that all non-noise
            clusters in that region should be covered by a polygon
        :param noiseLabel: label to associate with noise or None
        """

        # The constructor might seem bloated but it really mostly does input validation for the polygons
        coordinates = extractCoordinatesArray(coordinates)
        if isinstance(groundTruthPolygons, str):
            polygons: Sequence[Polygon] = gp.read_file(groundTruthPolygons).geometry.values
        elif isinstance(groundTruthPolygons, GeoDataFrame):
            polygons: Sequence[Polygon] = groundTruthPolygons.geometry.values
        else:
            polygons = groundTruthPolygons
        self.regionPolygon = polygons[0]
        self.noiseLabel = noiseLabel
        self.clusterPolygons = MultiPolygon(polygons[1:])
        self.noisePolygon = self.regionPolygon.difference(self.clusterPolygons)

        self.regionMultipoint = MultiPoint(coordinates).intersection(self.regionPolygon)
        if self.regionMultipoint.is_empty:
            raise Exception(f"The ground truth region contains no datapoints. "
                            f"This can happen if you have provided unsuitable coordinates")
        self.noiseMultipoint = self.regionMultipoint.intersection(self.noisePolygon)
        if self.noiseLabel is None and not self.noisePolygon.is_empty:
            raise Exception(f"No noise_label was provided but there is noise: {len(self.noiseMultipoint)} datapoints"
                            f"in annotated area do not belong to any cluster polygon")
        self.clustersMultipoints = []
        intermediatePolygon = Polygon()
        for i, clusterPolygon in enumerate(self.clusterPolygons, start=1):
            if not intermediatePolygon.intersection(clusterPolygon).is_empty:
                raise Exception(f"The polygons should be non-intersecting: polygon {i} intersects with previous polygons")
            intermediatePolygon = intermediatePolygon.union(clusterPolygon)
            clusterMultipoint = self.regionMultipoint.intersection(clusterPolygon)
            if clusterMultipoint.is_empty:
                raise Exception(f"The annotated cluster for polygon {i} is empty - check your data!")
            self.clustersMultipoints.append(clusterMultipoint)

    def toGeoDF(self, crs='epsg:3857', include_noise=True):
        """
        :return: GeoDataFrame with clusters as MultiPoint instance indexed by the clusters' identifiers
        """
        clusters = self.clustersMultipoints
        firstLabel = 0
        if self.noiseLabel is not None and include_noise:
            clusters = [self.noiseMultipoint] + clusters
            firstLabel = self.noiseLabel
        gdf = gp.GeoDataFrame({"geometry": clusters,
                               "identifier": list(range(firstLabel, firstLabel + len(clusters), 1))}, crs=crs)
        gdf.set_index("identifier", drop=True, inplace=True)
        return gdf

    def plot(self, includeNoise=True, **kwargs):
        """
        Plots the ground truth clusters

        :param includeNoise:
        :param kwargs:
        :return:
        """
        gdf = self.toGeoDF(include_noise=includeNoise)
        gdf["color"] = np.random.random(len(gdf))
        if includeNoise and self.noiseLabel is not None:
            gdf.loc[self.noiseLabel, "color"] = 0
        gdf.plot(column="color", **kwargs)

    def getCoordinatesLabels(self):
        """
        Extract cluster coordinates and labels as numpy arrays from the provided ground truth region and
        cluster polygons

        :return: tuple of arrays of the type (coordinates, labels)
        """
        coords, labels = [], []
        for row in self.toGeoDF(include_noise=True).itertuples():
            clusterMultipoint, label = row.geometry, row.Index
            coords += [[p.x, p.y] for p in clusterMultipoint]
            labels += [label] * len(clusterMultipoint)
        return np.array(coords), np.array(labels)



import logging
from typing import Sequence, Union, Optional

import torch

from .residualffn_modules import ResidualFeedForwardNetwork
from ...torch_base import VectorTorchModel, TorchVectorRegressionModel
from ...torch_opt import NNOptimiserParams
from ....normalisation import NormalisationMode

log: logging.Logger = logging.getLogger(__name__)


class ResidualFeedForwardNetworkTorchModel(VectorTorchModel):

    def __init__(self, cuda: bool, hiddenDims: Sequence[int], bottleneckDimensionFactor: float = 1, pDropout=None,
            useBatchNormalisation: bool = False) -> None:
        super().__init__(cuda=cuda)
        self.hiddenDims = hiddenDims
        self.bottleneckDimensionFactor = bottleneckDimensionFactor
        self.pDropout = pDropout
        self.useBatchNormalisation = useBatchNormalisation

    def createTorchModuleForDims(self, inputDim: int, outputDim: int) -> torch.nn.Module:
        return ResidualFeedForwardNetwork(inputDim, outputDim, self.hiddenDims, self.bottleneckDimensionFactor,
            pDropout=self.pDropout, useBatchNormalisation=self.useBatchNormalisation)


class ResidualFeedForwardNetworkVectorRegressionModel(TorchVectorRegressionModel):

    def __init__(self, hiddenDims: Sequence[int], bottleneckDimensionFactor: float = 1, cuda: bool = True, pDropout: Optional[float] = None,
            useBatchNormalisation: bool = False, normalisationMode: NormalisationMode = NormalisationMode.NONE,
            nnOptimiserParams: Union[NNOptimiserParams, dict, None] = None) -> None:
        super().__init__(ResidualFeedForwardNetworkTorchModel, [cuda, hiddenDims],
            dict(bottleneckDimensionFactor=bottleneckDimensionFactor, pDropout=pDropout, useBatchNormalisation=useBatchNormalisation),
            normalisationMode=normalisationMode, nnOptimiserParams=nnOptimiserParams)



import numpy as np
from typing_extensions import Protocol

from . import EuclideanClusterer


class SkLearnClustererProtocol(Protocol):
    """
    Only used for type hints, do not instantiate
    """
    def fit(self, x: np.ndarray): ...

    labels_: np.ndarray


class SkLearnEuclideanClusterer(EuclideanClusterer):
    """
    Wrapper around an sklearn-type clustering algorithm

    :param clusterer: a clusterer object compatible the sklearn API
    :param noiseLabel: label that is associated with the noise cluster or None
    :param minClusterSize: if not None, clusters below this size will be labeled as noise
    :param maxClusterSize: if not None, clusters above this size will be labeled as noise
    """

    def __init__(self, clusterer: SkLearnClustererProtocol, noiseLabel=-1,
             minClusterSize: int = None, maxClusterSize: int = None):
        super().__init__(noiseLabel=noiseLabel, minClusterSize=minClusterSize, maxClusterSize=maxClusterSize)
        self.clusterer = clusterer

    def _computeLabels(self, x: np.ndarray):
        self.clusterer.fit(x)
        return self.clusterer.labels_

    def __str__(self):
        return f"{super().__str__()}_{self.clusterer.__class__.__name__}"


from typing import Callable, Optional, Sequence

import torch
from torch import nn

from ...torch_base import MCDropoutCapableNNModule
from ....util.string import objectRepr


class MultiLayerPerceptron(MCDropoutCapableNNModule):
    def __init__(self, inputDim: float, outputDim: float, hiddenDims: Sequence[int],
            hidActivationFn: Callable[[torch.Tensor], torch.Tensor] = torch.sigmoid,
            outputActivationFn: Optional[Callable[[torch.Tensor], torch.Tensor]] = torch.sigmoid,
            pDropout: Optional[float] = None):
        super().__init__()
        self.inputDim = inputDim
        self.outputDim = outputDim
        self.hiddenDims = hiddenDims
        self.hidActivationFn = hidActivationFn
        self.outputActivationFn = outputActivationFn
        self.pDropout = pDropout
        self.layers = nn.ModuleList()
        if pDropout is not None:
            self.dropout = nn.Dropout(p=pDropout)
        else:
            self.dropout = None
        prevDim = inputDim
        for dim in [*hiddenDims, outputDim]:
            self.layers.append(nn.Linear(prevDim, dim))
            prevDim = dim

    def __str__(self):
        return objectRepr(self, dict(inputDim=self.inputDim, outputDim=self.outputDim, hiddenDims=self.hiddenDims,
            hidActivationFn=self.hidActivationFn.__name__ if self.hidActivationFn is not None else None,
            outputActivationFn=self.outputActivationFn.__name__ if self.outputActivationFn is not None else None,
            pDropout=self.pDropout))

    def forward(self, x):
        for i, layer in enumerate(self.layers):
            isLast = i+1 == len(self.layers)
            x = layer(x)
            if not isLast and self.dropout is not None:
                x = self.dropout(x)
            activation = self.hidActivationFn if not isLast else self.outputActivationFn
            if activation is not None:
                x = activation(x)
        return x



import functools
import logging
from typing import Optional, Sequence, Union, Any, Callable

from sklearn.preprocessing import MaxAbsScaler, StandardScaler, RobustScaler
import numpy as np
from typing_extensions import Protocol

log = logging.getLogger(__name__)

TransformableArray = Union[np.ndarray, Sequence[Sequence[Any]]]


def to2DArray(arr: TransformableArray) -> np.ndarray:
    if not isinstance(arr, np.ndarray):
        arr = np.array(arr)
    if len(arr.shape) != 2:
        raise ValueError(f"Got array of shape {arr.shape}; expected 2D array")
    return arr


class SkLearnTransformerProtocol(Protocol):
    def inverse_transform(self, arr: TransformableArray) -> np.ndarray:
        pass

    def transform(self, arr: TransformableArray) -> np.ndarray:
        pass

    def fit(self, arr: TransformableArray):
        pass


class ManualScaler:
    """
    A scaler whose parameters are not learnt from data but manually defined
    """
    def __init__(self, centre: Optional[float] = None, scale: Optional[float] = None):
        """
        :param centre: the value to subtract from all values (if any)
        :param scale: the value with which to scale all values (after removing the centre)
        """
        self.centre = centre if centre is not None else 0.0
        self.scale = scale if scale is not None else 1.0

    def fit(self, arr):
        pass

    def transform(self, arr: TransformableArray) -> np.ndarray:
        arr = to2DArray(arr)
        return (arr - self.centre) * self.scale

    def inverse_transform(self, arr: TransformableArray) -> np.ndarray:
        arr = to2DArray(arr)
        return (arr / self.scale) + self.centre


class SkLearnTransformerFactoryFactory:
    @staticmethod
    def MaxAbsScaler() -> Callable[[], MaxAbsScaler]:
        return MaxAbsScaler

    @staticmethod
    def StandardScaler(with_mean=True, with_std=True) -> Callable[[], StandardScaler]:
        return functools.partial(StandardScaler, with_mean=with_mean, with_std=with_std)

    @staticmethod
    def RobustScaler(quantile_range=(25, 75), with_scaling=True, with_centering=True) -> Callable[[], RobustScaler]:
        """
        :param quantile_range: a tuple (a, b) where a and b > a (both in range 0..100) are the percentiles which determine the scaling.
            Specifically, each value (after centering) is scaled with 1.0/(vb-va) where va and vb are the values corresponding to the
            percentiles a and b respectively, such that, in the symmetric case where va and vb are equally far from the centre,
            va will be transformed into -0.5 and vb into 0.5.
            In a uniformly distributed data set ranging from `min` to `max`, the default values of a=25 and b=75 will thus result in
            `min` being mapped to -1 and `max` being mapped to 1.
        :param with_scaling: whether to apply scaling based on quantile_range.
        :param with_centering: whether to apply centering by subtracting the median.
        :return: a function, which when called without any arguments, produces the respective RobustScaler instace.
        """
        return functools.partial(RobustScaler, quantile_range=quantile_range, with_scaling=with_scaling, with_centering=with_centering)

    @staticmethod
    def ManualScaler(centre: Optional[float] = None, scale: Optional[float] = None):
        return functools.partial(ManualScaler, centre=centre, scale=scale)


import logging

from ..tracking import TrackedExperiment

from clearml import Task

log = logging.getLogger(__name__)


# TODO: this is an initial working implementation, it should eventually be improved
class ClearMLExperiment(TrackedExperiment):
    def __init__(self, task: Task = None, projectName: str = None, taskName: str = None,
            additionalLoggingValuesDict=None):
        """

        :param task: instances of trains.Task
        :param projectName: only necessary if task is not provided
        :param taskName: only necessary if task is not provided
        :param additionalLoggingValuesDict:
        """
        if task is None:
            if projectName is None or taskName is None:
                raise ValueError("Either the trains task or the project name and task name have to be provided")
            self.task = Task.init(project_name=projectName, task_name=taskName, reuse_last_task_id=False)
        else:
            if projectName is not None:
                log.warning(
                    f"projectName parameter with value {projectName} passed even though task has been given, "
                    f"will ignore this parameter"
                )
            if taskName is not None:
                log.warning(
                    f"taskName parameter with value {taskName} passed even though task has been given, "
                    f"will ignore this parameter"
                )
            self.task = task
        self.logger = self.task.get_logger()
        super().__init__(additionalLoggingValuesDict=additionalLoggingValuesDict)

    def _trackValues(self, valuesDict):
        self.task.connect(valuesDict)
